import bisect
import numpy
import time
from .      import __version__
from .space import SpaceList, CfDict, Coordinate
from .utils import equals, array_allclose


_global_attributes = set(('comment',
                          'Conventions',
                          'history',
                          'institution',
                          'references',
                          'source',
                          'title',
                          ))

class AggregateError(Exception):
    '''

An exception that indicates that two fields can not be aggregated.
'''
    pass
#--- End: class


def aggregate(fields, rtol=None, atol=None, Ndim=True,
              verbose=False, messages=False):
    '''

'''
    aggregated = SpaceList()
        
    standard_name_group = {}
    for f in fields:
        try:
            standard_name = f.standard_name
        except AttributeError:
            # This field has no standard name, so it can not be
            # aggregated. Put it straight into the output list.
            aggregated.append(f)
        else:
            try:
                # Append this field to the group of fields with this
                # standard name
                standard_name_group[standard_name].append(f)
            except KeyError:
                # Initialize the group of fields with this standard name
                standard_name_group[standard_name] = SpaceList(f)
    #--- End: for    
    
    for f in standard_name_group.values():
        original_len_f = len(f)
        M = 0
        while 1:
#            print 'round',M,'of', f[0].standard_name
#            print repr(f)
            len_f = len(f)
            i = 0
            while i < len(f):
#                print 'len(f) =', len(f)
                j = i + 1
                while j < len(f):
                    try:
#                        print 'i,j=',i,j, '(', len_f,') M=', M
                        f[i] = aggregate_2_fields(f[i], f[j], 
                                                  rtol=rtol, atol=atol, 
                                                  Ndim=Ndim)
                        # Successfully aggregated fields f[i] and f[j]
                        # into f[i], so get rid of f[j]
                        f.pop(j)
                    except AggregateError as error:
                        # Could not aggregate fields f[i] and f[j]
                        j += 1
                        if messages:
                            print "Can't aggregate %s: %s" % (f[0].standard_name, error)
                #--- End: while 
                i += 1
            #--- End: while 
            if not Ndim or len(f) == len_f:           
                break
            len_f = len(f)
            M += 1
        #--- End: while
        aggregated.extend(f)

        if verbose:
            print "Aggregated %d '%s' field%s to %d field%s" % \
                (original_len_f, f[0].standard_name, 
                 ('s' if original_len_f>1 else ''),                 
                 len_f, ('s' if len_f>1 else ''))
    #--- End: for

    # Update input fields in-place
    fields[:] = aggregated  

    fields.setpub('history', '%s Aggregated by cf-python v%s' % \
                      (time.strftime('%Y-%m-%d %H:%M:%S UTC', time.gmtime()),
                       __version__))
                  
#--- End: def

def _equivalent_data(var0, var1, 
                     var0_dimensions, var1_dimensions, 
                     match_for_dim1, rtol=None, atol=None):
    '''
    
'''
    # ----------------------------------------------------------------
    # Check that the variables' data sizes are the same
    # ----------------------------------------------------------------

    size0 = var0.size
    size1 = var1.size
    if size0 != size1:
        raise ValueError, "Variables have different sizes"

    # ----------------------------------------------------------------
    # Compare the variables' data, after making them compatible
    # ----------------------------------------------------------------
    x = var0._data.concatenate()._list[0]
    y = var1._data.concatenate()._list[0]

    new_order  = []
    full_slice = [slice(None)] * y.ndim

    if size0 == 1:
        y1 = y
    else:
        # Reorder y so that its axes are in the same order as x
        for dim1 in var1_dimensions:
            new_order.append(var0_dimensions.index(match_for_dim1[dim1]))       
        y1 = numpy.transpose(y, axes=new_order)

        if x.shape != y.shape:
            raise ValueError, "Variables have different shapes"

        # Reverse any of y's axes thst increase in the opposite
        # direction to the same axis in x.
        full_slice = [slice(None)] * y1.ndim
        for i in xrange(x.ndim):
            dir0 = var0._data.direction[var0._data.order[i]]
            dir1 = var0._data.direction[var1._data.order[new_order[i]]]
            if dir0 != dir1:
                full_slice[i] = slice(None, None, -1)
        #--- End: for  
        y1 = y1[tuple(full_slice)]
    #--- End: if  

    if not array_allclose(x, y1, rtol=rtol, atol=atol):
        raise ValueError, "Variables have different data values"

    # ----------------------------------------------------------------
    # Compare the variables' bounds, after making them compatible
    # ----------------------------------------------------------------
    if isinstance(var0, Coordinate):

        try:
            x = var0.bounds._data.concatenate()._list[0]
            y = var1.bounds._data.concatenate()._list[0]
        except AttributeError: 
            pass
        else:
            y1 = numpy.transpose(y, axes=new_order+[y.ndim-1])
            dir0 = var0.bounds._data.direction['bounds']
            dir1 = var0.bounds._data.direction['bounds']
            if dir0 != dir1:
                full_slice.append(slice(None, None, -1))
                y1 = y1[tuple(full_slice)]
            else:
                full_slice.append(slice(None))
            
            if not array_allclose(x, y1, rtol=rtol, atol=atol):
                raise ValueError, "Coordinates have different bounds values"
        #--- End: try
    #--- End: if

    # Still here? Then these two variables have equivalent data
    return True
#-- End: def

def _equivalent_properties(var0, var1, rtol=None, atol=None):
    '''
    
Test for equivalent coordinate properties.

Returns True if the two coordinates have equivalent properties and
raises a ValueError otherwise.
'''
    # Test Units. Note that this is currently too strict - non-equal
    # but equivalent Units are currently not allowed, but they should be.
    try:
        if not var0._Units.equals(var1._Units, rtol=rtol, atol=atol):
            raise ValueError, \
"Different Units"
    except AttributeError:
        # At least one variable doesn't have Units
        raise ValueError, \
"Missing Units"

    cfatts0 = set(var0.pubattrs) - set(('standard_name', 'units', 'calendar'))
    cfatts1 = set(var1.pubattrs) - set(('standard_name', 'units', 'calendar'))

    if cfatts0 != cfatts1:
        raise ValueError, \
"Different sets of attributes"

    for attr in cfatts0:
        if not equals(var0.getpub(attr), var1.getpub(attr),
                      rtol=rtol, atol=atol):
            raise ValueError, \
"Different values for attribute '%(attr)s'" % locals()
    #-- End: for

    # Check consistent existence of bounds or transform attributes
    for attr in ('bounds', 'transform'): # dch false clim?
        try:
            getattr(var0, attr)
        except AttributeError:
            try:
                getattr(var1, attr)
                raise ValueError, \
"Missing attribute '%(attr)s'" % locals()
            except AttributeError:
                pass
        else:
            try:
                getattr(var1, attr)
            except AttributeError:
                raise ValueError, \
"Missing attribute '%(attr)s'" % locals()
        #--- End: try
    #--- End: for

    try:
        if not var0.transform.equals(var1.transform, rtol=rtol, atol=atol):
            raise ValueError, \
"Different transforms"
    except AttributeError:
       pass
  
    # Check consistenty of climatology
    try:
        climatology0 = var0.climatology
    except AttributeError:
        try:
            if var1.climatology:
                raise ValueError, \
"Different climatology"
        except AttributeError:
            # Neither variable has a climatology attribute 
            pass
    else:
        try:
            climatology1 = var1.climatology
        except AttributeError:
            if climatology0:
                raise ValueError, \
"Different climatology"
        else:
            # Both variables have a climatology attribute 
            if bool(climatology0) != bool(climatology1):
                raise ValueError, \
"Different climatology"        
    #--- End: try

    # Still here?
    return True
#--- End: def

def aggregate_2_fields(field0, field1, rtol=None, atol=None, Ndim=False):

#- Both fields have the same standard_name, etc.

#- Each dimension in each space has at least one scalar or 1-d
#  coordinate (dimension or auxiliary) and each such coordinate's
#  standard_name is unique within its space.

    if field0.ndim < field1.ndim:
        field0, field1 = field1, field0



    space0 = field0.grid
    space1 = field1.grid

    # Raise an exception for missing functionality
    for attr in ('ancillary_variables', 'flag_values', 'flag_masks', 'flag_meanings'):
        for field in (field0, field1):
            try:
                getattr(field, attr)
                raise AggregateError, \
"Can't aggregate %(attr)s attributes, yet." % locals()
            except AttributeError:
                pass
    #--- End: for

    # Test Units. Note that this is currently too strict - non-equal
    # but equivalent Units are currently not allowed
    try:
        if not field0._Units.equals(field1._Units, rtol=rtol, atol=atol):
            raise AggregateError, \
"Different field Units"
    except AttributeError:
        # At least one field doesn't have Units
        raise AggregateError, \
"Missing field Units"

    if len(space0.dimension_sizes) != len(space1.dimension_sizes):
        raise AggregateError, \
"Different ranks for spaces"
         
    cfatts0 = (set(field0.pubattrs) - 
               _global_attributes -
               set(('cell_methods', 'standard_name', 'units', 'calendar'))
               )
    cfatts1 = (set(field1.pubattrs) -
               _global_attributes -
               set(('cell_methods', 'standard_name', 'units', 'calendar'))
               )

    if cfatts0 != cfatts1:
        raise AggregateError, \
"Different sets of field attributes"

    for attr in cfatts0:
        if not equals(field0.getpub(attr), field1.getpub(attr),
                      rtol=rtol, atol=atol):
            raise AggregateError, \
"Different values for field attribute '%(attr)s'" % locals()

    # Each dimension in each space has at least one scalar or 1-d
    # coordinate (dimension or auxiliary) and each such coordinate's
    # standard_name is unique within its space.
    meta0 = CfDict()
    meta1 = CfDict()

    meta  = meta0
    space = space0
    field = field0
    for n in (0, 1):

        all_standard_names     = []
        meta.nd_coords         = []
        meta.nd_standard_names = []

        for key in space:
            coord = space[key]            
            
            try:
                standard_name = coord.standard_name
            except AttributeError:
                raise AggregateError, \
"Coordinate or cell measure does not have a standard_name"

            all_standard_names.append(standard_name)

            try:
                if coord.ndim <= 1:
                    dim = space.dimensions[key][0]                    
                else:
                    meta.nd_coords.append(key)
                    meta.nd_standard_names.append(standard_name)
                    continue
            except AttributeError:
                raise AggregateError, \
"Coordinate or cell measure has no data"

            # Still here?
            try:
                meta[dim]['1d_coords'].append(key)
                meta[dim]['1d_standard_names'].append(standard_name)
            except KeyError:
                meta[dim] = {'1d_coords'        : [key],
                             '1d_standard_names': [standard_name]}
        #--- End: for

        if len(all_standard_names) > len(set(all_standard_names)):
            raise AggregateError, \
"Space components in one field share a standard_name"

        # Check that every dimension of the space has at least one
        # dimension coordinate or 1-d auxiliary coordinate
        for dim in space.dimension_sizes:
            try:
                meta[dim]
            except KeyError:
                raise AggregateError, \
"Dimension has no scalar or 1-d coordinates"
            
            # Create a set of each dimension's 1-d coordinates' standard
            # names
            meta[dim]['set_1d_standard_names'] = \
                set(meta[dim]['1d_standard_names'])
        #--- End: for

        # Change space ready for the next loop iteration
        meta  = meta1
        space = space1
        field = field1
    #--- End: for

    if len(meta0.nd_coords) != len(meta1.nd_coords):
        raise AggregateError, \
"Different numbers of N-d coordinate/cell measures"

    # ----------------------------------------------------------------
    # Check the cell methods. We do this after ascertaining that every
    # coordinate has a standard name so we can be sure that the 'name'
    # keys are comparable
    # ----------------------------------------------------------------
    try:
        cell_method0 = field0.cell_methods
    except AttributeError:
        try:
            field1.cell_methods
            raise AggregateError, \
"Different cell methods"
        except AttributeError:
            # Neither field has cell methods
            pass
    else:
        try:
            cell_method1 = field1.cell_methods
        except AttributeError:
            raise AggregateError, \
"Different cell methods"

        # ------------------------------------------------------------
        # Both fields have cell methods, so compare them method-wise
        # ------------------------------------------------------------
        if len(cell_method0) != len(cell_method1):
            raise AggregateError, \
"Different cell methods"

        for cm0, cm1 in zip(cell_method0, cell_method1):
            keys0 = set(cm0) - set(('dim',))
            keys1 = set(cm1) - set(('dim',))
        
            if keys0 != keys1:
                raise AggregateError, \
"Different cell methods"

            for key in keys0:
                if not equals(cm0[key], cm1[key], rtol=rtol, atol=atol):
                    raise AggregateError, \
"Different cell methods"
        #--- End: for
    #--- End: try

    # ================================================================
    # For each dimension in one of the spaces, there is a 'matching
    # dimension' in the other space with exactly the same list of
    # scalar or 1-d coordinate (dimension or auxiliary)
    # standard_names, although a standard_name may belong to different
    # types of coordinate in the two spaces. Call these coordinates a
    # pair of 'matching coordinates' for the matching dimensions
    # ================================================================
    matching_dimensions = {}
    match_for_dim1      = {'bounds': 'bounds'}
    dims1 = space1.dimension_sizes.keys()

    for dim0 in space0.dimension_sizes:

        # Make sure name 'dim1' is not defined
        try:
            del dim1
        except NameError:
            pass

        standard_names0 = meta0[dim0]['1d_standard_names'][0]
        for i in xrange(len(dims1)):
            dim = dims1[i]
            if standard_names0 in meta1[dim]['set_1d_standard_names']:
                # Found a potentially matching dimension
                dim1 = dim
                dims1.pop(i)
                break
        #--- End: for

        try:
            matched_dims = (dim0, dim1)
        except NameError:
            # dim1 is not defined
            raise AggregateError, \
"Not all dimensions can be matched"

        # Add to the dictionaries which, given a dimension from one
        # field, easily return its matching dimension from the other
        # field. This is used to reorder variables from field1 when
        # checking that they have equivalent values and for inserting
        # one variable into another
        match_for_dim1[dim1] = dim0

        # Now we have a pair of potentially matching dimensions, check
        # to see if it really is; and if it is then record its
        # matching coordinates.
        matching_dimensions[matched_dims] = []
        for key0, sn0 in zip(meta0[dim0]['1d_coords'], 
                             meta0[dim0]['1d_standard_names']):
            try:
                i = meta1[dim1]['1d_standard_names'].index(sn0)
            except ValueError:
                raise AggregateError, \
"Dimensions do not share the same set of standard names"

            key1 = meta1[dim1]['1d_coords'][i]

            try:
                _equivalent_properties(space0[key0], space1[key1],
                                       rtol=rtol, atol=atol)
            except ValueError as error:
                # space0[key0] has no matching coordinate in space1
                raise AggregateError, \
"%s has no match: %s" % (repr(space0[key0]), error)

            matching_dimensions[matched_dims].append((key0, key1))            
        #--- End: for
    #--- End: for

#- Paired nd auxes
    matching_nd_auxiliaries = []
    for key0, sn0 in zip(meta0.nd_coords, meta0.nd_standard_names):
   
        try:
            i = meta1.nd_standard_names.index(sn0)
        except ValueError:
            # The two fields' N-d coordinates/cell measures do not
            #  share the same set of standard names
            raise AggregateError, \
"%s has no match" % repr(space0[key0])

        key1 = meta1.nd_coords[i]
  
        # Check that potentially matching N-d coordinates/cell
        # measures span equivalent dimensions
        dims0 = space0.dimensions[key0][:]
        dims1 = space1.dimensions[key1][:]
        if len(dims0) != len(dims1):
            # N-d coordinates/cell measures have different ranks
            raise AggregateError, \
"N-d coordinate/cell measure has no matching coordinate"
    
        for dim0, dim1 in matching_dimensions:
            try:
                dims0.remove(dim0)
            except ValueError:
                pass
            else:
                try:
                    dims1.remove(dim1)
                except ValueError:
                    raise AggregateError, \
"N-d coordinate/cell measure has no matching coordinate"
    
                else:
                    if not dims0:
                         break
        #--- End: for        
    
        # Check that potentially matching N-d coordinates/cell
        # measures have equivalent properties
        try:
            _equivalent_properties(space0[key0], space1[key1],
                                   rtol=rtol, atol=atol)
        except ValueError:
            raise AggregateError, \
"N-d coordinate/cell measure has no matching coordinate"
    
        matching_nd_auxiliaries.append((key0, key1))
    #--- End: for

    # ================================================================
    # There is EXACTLY ONE pair of matching dimensions for which at
    # least one pair of matching 1-d coordinates do NOT share the same
    # values for their arrays and bounds. Call this dimension the
    # 'aggregating dimension' in each field.
    # ================================================================
    for matched_dims, matching_coords in matching_dimensions.iteritems():

        got_dim = False
        for key0, key1 in matching_coords:
           if key0.startswith('d') or key1.startswith('d'):
               got_dim = True
               break
        #--- End: for

        if not got_dim:
            try:
                aggregating 
                raise AggregateError, \
"More than one pair of dimensions have 1-d coordinates with different values"
            except NameError:
                # The name 'aggregating' is not already defined, this
                # pair of matching dimensions comprise the aggregating
                # dimensions, assuming we don't find another pair with
                # unequal data.
                aggregating = matched_dims
                continue
        #--- End: if

        # Still here?
        for key0, key1 in matching_coords:
            try:
                _equivalent_data(space0[key0], space1[key1],
                                 space0.dimensions[key0], space1.dimensions[key1],
                                 match_for_dim1, rtol=rtol, atol=atol)
                # Still here? Then these coordinates have equivalent
                # values
                continue                
            except ValueError:
                pass

            # Still here? Then these coordinates have non-equivalent
            # values and so this pair of matching dimensions could
            # comprise the aggregating dimensions
            try:
                aggregating
                raise AggregateError, \
"More than one pair of dimensions have 1-d coordinates with different values"
            except NameError:
                # The name 'aggregating' is not already defined, so
                # this pair of matching dimensions comprise the
                # aggregating dimensions, assuming we don't find
                # another pair with unequal data.
                aggregating  = matched_dims
                break
        #--- End: for
    #--- End: for

    # ----------------------------------------------------------------
    # Find which pairs of matching coordinates/cell measures (1-d and
    # Nd) span their aggregating dimension and which do not. Also
    # find, for each field, the key of its aggregating dimension.
    # ----------------------------------------------------------------
    try:        
        spanning_1d_coordinates = matching_dimensions[aggregating]
    except NameError:
        # The name 'aggregating' is not defined because each pair of
        # matching coordinates has the same data values
        raise AggregateError, \
"No pair of aggregating dimensions"

    key0, key1 = spanning_1d_coordinates[0]
    adim_key0 = space0.dimensions[key0][0]
    adim_key1 = space1.dimensions[key1][0]

    spanning_nd_auxiliaries     = []
    non_spanning_nd_auxiliaries = []
    for key0, key1 in matching_nd_auxiliaries:
        if adim_key0 in space0.dimensions[key0]:            
            spanning_nd_auxiliaries.append((key0, key1))
        else:
            non_spanning_nd_auxiliaries.append((key0, key1))

    # ----------------------------------------------------------------
    # Check that both or neither aggregating dimensions have a
    # dimension coordinate. (Note that this is stricter than
    # necessary, but means that, for now, we definitely don't have to
    # sort one of the fields along its aggregating dimension.)
    # ----------------------------------------------------------------
    for key0, key1 in spanning_1d_coordinates:
        if key0.startswith('d') and key1.startswith('d'):
            # Found a pair of dimension coordinates
            dim_coords = (key0, key1)
            break

        if key0.startswith('d') or key1.startswith('d'):
            raise AggregateError, \
"Can not (currently) join a dimension coordinate with an auxiliary coordinate"
    #--- End: for

  
    # ================================================================
    # If either or both aggregating dimensions have a dimension
    # coordinate then there are no values for it and its matching
    # coordinate which appear in both spaces (if they run in different
    # senses or their ranges overlap, the aggregated dimension will
    # need to be sorted). [All other coords are concatenated.]
    # ================================================================

    # ----------------------------------------------------------------
    # Find the insertion indices
    # ----------------------------------------------------------------
    try:
        dim_coords
    except NameError:
        # ------------------------------------------------------------
        # Neither field's aggregating dimension has a dimension
        # coordinate (because name 'dim_coord' is not
        # defined). Therefore the fields will be simply concatenated
        # along the aggregating dimension
        # ------------------------------------------------------------
        indices = len(field0._data)
    else:        
        # ------------------------------------------------------------
        # The aggregating dimensions of both fields have a dimension
        # coordinate. Therefore we need to find the insertion index in
        # field0 for each data element in field1.
        # ------------------------------------------------------------
        dim0 = space0[dim_coords[0]]
        dim1 = space1[dim_coords[1]]

        dim0._data.conform()
        dim1._data.conform()

        mins0 = []
        maxs0 = []
        for array in dim0._data._list:
            try:
                first = numpy.asscalar(array)
                mins0.append(first)
                maxs0.append(first)
            except ValueError:
                size       = array.size
                first_last = array[slice(0, size, size-1)]
                mins0.append(min(first_last))
                maxs0.append(max(first_last))
        #--- End: for

        # Set the reversed_ flag to signify whether or not we had to
        # reverse the mins0 and maxs0 lists in order to get them into
        # ascending sequences.
        try:
            reversed_ = not dim0._data.direction.values()[0]
        except AttributeError:
            reversed_ = not dim0._data.direction
        if reversed_:
            mins0.reverse()
            maxs0.reverse()
        
        indices = []
        for i, array in enumerate(dim1._data._list):
            try:
                first = numpy.asscalar(array)
                min1  = first
                max1  = first
            except ValueError:
                size       = array.size
                first_last = array[slice(0, size, size-1)]
                min1       = min(first_last)
                max1       = max(first_last)

            index = bisect.bisect(mins0, min1)

            if reversed_:
                indices.append(len(mins0)-index)
            else:
                indices.append(index)

            # Update mins0 and maxs0
            mins0.insert(index, min1)
            maxs0.insert(index, max1)

            # Check for overlapping dimension coordinates (not allowed
            # for now). This check also serves to check for any shared
            # dimension coordinate values.
            if index == len(maxs0)-1:
                if min1 <= maxs0[index-1]:
                    raise AggregateError, \
"Overlapping or equal dimension coordinate values"
            elif index == 0:
                if max1 >= mins0[index+1]:
                    raise AggregateError, \
"Overlapping or equal dimension coordinate values"
            else:
                 if min1 <= maxs0[index-1] or max1 >= mins0[index+1]:
                     raise AggregateError, \
"Overlapping or equal dimension coordinate values"
            #--- End: if

            # Check for overlapping dimension coordinate bounds. This
            # is a bit strict for now
            try:               
                bounds1 = dim1.bounds._data._list[i]
            except AttributeError:
                # Dimension coordinate knows no bounds
                pass
            else:
                # Dimension coordinate has bounds
                if index == len(maxs0)-1:
                    if numpy.amin(bounds1) <= maxs0[index-1]:
                        raise AggregateError, \
"Overlapping (ish) dimension coordinate bounds"
                elif index == 0:
                    if numpy.amax(bounds1) >= mins0[index+1]:
                        raise AggregateError, \
"Overlapping (ish) dimension coordinate bounds"
                else:  
                    if (numpy.amin(bounds1) <= maxs0[index-1] or
                        numpy.amax(bounds1) >= mins0[index+1]):
                        raise AggregateError, \
"Overlapping (ish) dimension coordinate bounds"
            #--- End: try
        #--- End: for
    #--- End: if

    # ================================================================
    # Check that N-d auxiliary coordinates/cell measures which don't
    # span the aggregating dimension have the same data values
    # ================================================================
    for key0, key1 in non_spanning_nd_auxiliaries:
        try:
            _equivalent_data(space0[key0], space1[key1],
                             space0.dimensions[key0], space1.dimensions[key1],
                             match_for_dim1, rtol=rtol, atol=atol)            
        except ValueError:
            raise AggregateError, \
"N-d variables which do not span the aggregating dimension have different values"
    #--- End: for

    # ----------------------------------------------------------------
    # Check that all coordinates/field data/cell measures which span
    # the aggregating dimension are not already aggregated across a
    # different dimension and that, within each field, they all have
    # the same number of data elements. (Note that we're not checking
    # that each data element has an equivalent aggregating dimension
    # size. I'm assuming that this will be the case. But is this a
    # reasonable assumption?)
    # ----------------------------------------------------------------
    for adim_key, field in zip((adim_key0, adim_key1),
                               (field0   , field1)):

        try:
            adim = field.grid.dimensions['data'].index(adim_key)
        except ValueError:
            # Field's data doesn't span the aggregating dimension
            try:
                field._data.adim
                # Field is aggregated along a dimension other than the
                # aggregating dimension 
                if not Ndim:
                    raise AggregateError, \
"Field's data are already aggregated along a different dimension" 
                else:
                    field._concatenate_data()
            except AttributeError:
                # Field's data is not aggregated along any dimension
                pass
        else:
            # Field's data does span the aggregating dimension
            try:            
                if field._data.adim != adim:
                    # Field is aggregated along a dimension other than
                    # the aggregating dimension
                    if not Ndim:
                        raise AggregateError, \
"Field's data are already aggregated along a different dimension"
                    else:
                        field._concatenate_data()
            except AttributeError:
                # Field's data is not aggregated along any dimension
                pass
        #--- End: try
    #--- End: for
  
    adim_key   = adim_key0
    n_elements = len(field0._data)
    for i, space in enumerate((space0, space1)):
    
        # Check spanning 1-d coordinates
        for key_pair in spanning_1d_coordinates:
            if len(space[key_pair[i]]._data) != n_elements:
                raise AggregateError, \
"1-d coordinate has the wrong number of data elements %d %d" % (len(space[key_pair[i]]._data),n_elements)
        #--- End: for

        # Check spanning N-d coordinates/cell measures
        for key_pair in spanning_nd_auxiliaries:
            key  = key_pair[i]
            if len(space[key]._data) != n_elements:
                raise AggregateError, \
"N-d coordinate/cell measures has the wrong number of data elements"

#            adim = space.dimensions[key].index(adim_key)
#            try:            
#                if space[key]._data.adim != adim:
#                    if not Ndim:
#                        raise AggregateError, \
#"N-d coordinate/cell measures is already aggregated along a different dimension"
#                    else:
#                        space[key]._concatenate_data()
#            except AttributeError:
#                pass
        #--- End: for

        adim_key   = adim_key1
        n_elements = len(field1._data)
    #--- End: for

    # ================================================================
    # All of the aggregation tests have been passed and so field1 will
    # now be aggregated into field0 *in place*. Field1 might also be
    # changed *in place*.
    # 
    # There are no more trapped AggregateError exceptions from this
    # point onwards, as it is assumed that the fields have passed all
    # of the aggregation checeks and are therefore aggregatable.
    # ================================================================

    # Get 1-d coordinates ready for aggregation
    for key_pair in spanning_1d_coordinates:
        c = space0[key_pair[0]]
        try:
            c._data.adim
            continue
        except AttributeError:                
            pass

        c._data.adim = 0                            
        try:
            c.bounds._data.adim = 0
        except AttributeError:
            pass

        if not c._data.shape:
            c._data.shape     = (1,)
            c._data.order     = [adim_key0]
            c._data.direction = {adim_key0: c._data.direction}
            try:
                b = c.bounds._data
            except AttributeError:
                pass
            else:
                b.shape               = (1, 2)
                b.order.insert(0, adim_key0)
                b.direction[adim_key0] = b.direction['bounds']
        #--- End: if
    #--- End: for

#    adim_key = adim_key0
#    for i, space in enumerate((space0, space1)):
#        for key_pair in spanning_1d_coordinates:
#            c = space[key_pair[i]]
#            try:
#                c._data.adim
#                continue
#            except AttributeError:                
#                pass
#
#            c._data.adim = 0                            
#            try:
#                c.bounds._data.adim = 0
#            except AttributeError:
#                pass
#
#            if not c._data.shape:
#                c._data.shape     = (1,)
#                c._data.order     = [adim_key]
#                c._data.direction = {adim_key: c._data.direction}
#                try:
#                    b = c.bounds._data
#                except AttributeError:
#                    pass
#                else:
#                    b.shape               = (1, 2)
#                    b.order.insert(0, adim_key)
#                    b.direction[adim_key] = b.direction['bounds']
#            #--- End: if
#        #--- End: for
#        adim_key = adim_key1
#    #--- End: for

    # Get N-d coordinates and cell measures ready for aggregation
    adim_key = adim_key0
    for i, space in enumerate((space0, space1)):
        for key_pair in spanning_nd_auxiliaries:
            key = key_pair[i]
            c   = space[key]
            try:  
                c._data.adim
            except AttributeError:
                c._data.adim = space.dimensions[key].index(adim_key)            
                try:
                    c.bounds._data.adim = c._data.adim
                except AttributeError:
                    pass
            #--- End: try
        #--- End: for
        adim_key = adim_key1
    #--- End: for

    # Get the data ready for aggregation
    try:            
        field0._data.adim
    except AttributeError:
        try:
            # The field's data spans the aggregating dimension
            field0._data.adim = space0.dimensions['data'].index(adim_key0)
        except ValueError:
            # The field's data doesn't currently span the
            # aggregating dimension, so it needs to be inserted
            field0.grid.dimensions['data'].insert(0, adim_key0)

            field0._data.adim  = 0
            field0._data.shape = (1,) + field0._data.shape
            field0._data.order.insert(0, adim_key0)
            try:
                field0._data.direction.update(space0[adim_key0]._data.direction)
            except AttributeError:
                # Field's data is currently a scalar
                field0._data.direction = field0.grid[adim_key0]._data.direction.copy()
        #--- End: try
    #--- End: try

#   # Get the data ready for aggregation
#   for adim_key, field in zip((adim_key0, adim_key1),
#                              (field0   , field1   )):
#       try:            
#           field._data.adim
#       except AttributeError:
#           try:
#               # The field's data spans the aggregating dimension
#               field._data.adim = field.grid.dimensions['data'].index(adim_key)
#           except ValueError:
#               # The field's data doesn't currently span the
#               # aggregating dimension, so it needs to be inserted
#               field.grid.dimensions['data'].insert(0, adim_key)
#
#               field._data.adim  = 0
#               field._data.shape = (1,) + field._data.shape
#               field._data.order.insert(0, adim_key)
#               try:
#                   field._data.direction.update(field.grid[adim_key]._data.direction)
#               except TypeError:
#                   field._data.direction[adim_key] = field.grid[adim_key]._data.direction
#               except AttributeError:
#                   # Field's data is currently a scalar
#                   field._data.direction = field.grid[adim_key]._data.direction.copy()
#           #--- End: try
#       #--- End: try
#   #--- End: for

    # Insert 1-d coordinates/cell measures which span the aggregating
    # dimension
    for key0, key1 in spanning_1d_coordinates:
        space0[key0]._data_insert(indices, space1[key1],
                                  space0.dimensions[key0], space1.dimensions[key1],
                                  match_for_dim1)
        # Update dimension_sizes
        space0.dimension_sizes[adim_key0] = space0[key0].size

    # Insert N-d coordinates/cell measures which span the aggregating
    # dimension
    for key0, key1 in spanning_nd_auxiliaries:
        space0[key0]._data_insert(indices, space1[key1],
                                  space0.dimensions[key0], space1.dimensions[key1],
                                  match_for_dim1)
 
    # Insert data
    field0._data_insert(indices, field1,
                        space0.dimensions['data'], space1.dimensions['data'],
                        match_for_dim1)
    
    return field0
#--- End: def
