#!/usr/bin/python

import numpy
import re
import sys
import copy

numpy.seterr(all='raise')

def collapse(spaces, collapse, wtype={}, infer_bounds=False,
             weights=None, method='amean'):
    '''s is the space to collapse.

    Unless a weights array is provided (see below), weights for the
    collapse dimensions given by collapse and wtype will be calculated
    from each input space.

    collapse is an optional list of dimensions to collapse,
    e.g. ['latitude', 'longitude']. Names of coordinates implied by
    formula_terms may be given, e.g. ['pressure']. If a named collapse
    dimension has formula_terms attribute, then the weights will be
    calculated according to the virtual coordinate. Dimension names
    may be unabiguous abbreviations, e.g. ['lat','lon']. Dimension
    identifiers may be given in the event of ambiguous or missing
    coordinates, e.g. ['dim2'].

    wtype is an optional dictionary which relates collapse dimensions
    to specific weighting types, e.g.

        {'lat' : 'default', 
         'long': 'linear',
         'time': [1,2,1],
         'regi': myfunc,
         'dim4': 'equal'}

    Dimensions specified in the collapse list and not in the wtype
    dictionary are given a weighting type of 'default'. Dimensions
    specified in both the collapse list and the wtype dictionary are
    given weighting types from the latter. Valid weighting type
    strings are 'default', 'equal', 'linear', 'latitude_area'. A
    weighting type may also be a one dimensionsal array of weights or
    a user-defined function which returns a one dimensionsal array of
    weights.

    A weighting type of 'default' requires that the calc_weights
    function know what the default is for that dimension.

    infer_bounds determines whether cell boundaries should be inferred
    from cell coordinate values if there is no bounds array.

    weights is an optional cf variable containing the weights for the
    collapse. If this is provided, no weights will be calculated.

    method is the collapse method, currently one of:    
        'biased_sd'   : Biased standard deviation
        'unbiased_sd' : Unbiased standard deviation
        'biased_var'  : Biased variance
        'unbiased_var': Unbiased variance
        'amean'       : Arithmetic mean
        'gmean'       : Geometric mean
        'hmean'       : Harmonic mean    


    Example calls:

    >>> collapsed = collpase(spaces, ['lon','lat'])
    >>> collapsed = collpase(spaces, ['lon','lat'], method='unbiased_sd')
    >>> collapsed = collpase(spaces, {'lat': 'linear'}, method='amean')
    >>> collapsed = collpase(spaces, ['depth','lat'], {'lat': 'linear'})
    >>> collapsed = collpase(spaces, ['lat'], {depth=[0.5,1,0.5],'lat': 'default'})
    >>> cos_weights = lambda s,d,coord: numpy.array([cos(x) for x in s.coord[coord])
    >>> collapsed = collpase(spaces, ['depth'], {'lat': cos_weights})
                            
    '''

    # Initialize output list of spaces
    spaces_out = []
    
    # Make sure we have an input list of spaces
    if isinstance(spaces, cf):
        spaces = [spaces]

    # ----------------------------------------------------------
    # Create a dictionary of which functions need to be
    # defined. E.g. if 'biased_sd' is requested, then we need a
    # functions for 'amean', 'biased_var' and 'biased_sd'
    # ----------------------------------------------------------
    methods = {'biased_sd'   : False, # Biased standard deviation
               'unbiased_sd' : False, # Unbiased standard deviation
               'biased_var'  : False, # Biased variance
               'unbiased_var': False, # Unbiased variance
               'amean'       : False, # Arithmetic mean
               'gmean'       : False, # Geometric mean
               'hmean'       : False, # Harmonic mean
               }
    
    if method not in methods:
        error('ERROR: Unknown collapse method: '+method)

    methods[method] = True
    
    # Set dependencies between collapse methods (the order of
    # these statements matters)
    if methods['biased_sd']   : methods['biased_var']   = True
    if methods['unbiased_sd'] : methods['unbiased_var'] = True
    if methods['biased_var']  : methods['amean']        = True
    if methods['unbiased_var']: methods['amean']        = True
    
    # ----------------------------------------------------------------
    # Define the collapse function: fold(s, w, sum_w, n) where s = the
    # data to collapse, w = the weights, sum_w = sum of non-zero
    # weights, n = number of non-zero weights. The order of the
    # definitions matters.
    # ----------------------------------------------------------------
    if methods['amean']:
        # Arithmetic mean
        fold = lambda s, w, sum_w, n=None: \
            (s * w).sum() / sum_w
        amean = fold
        
    if methods['gmean']:
        # Geometric mean
        from math import exp
        fold = lambda s, w, sum_w, n=None: \
            exp((numpy.log(s) * w).sum() / sum_w)
        
    if methods['hmean']:
        # Harmonic mean
        fold = lambda s, w, sum_w, n=None:\
            sum_w / (w/s).sum()
        
    if methods['biased_var']:
        # Biased variance (requires amean)
        fold = lambda s, w, sum_w, n=None: \
            (w*(s-amean(s, w, sum_w))**2).sum() / sum_w
        biased_var = fold
        
    if methods['unbiased_var']:
        # Unbiased variance (requires amean)
        sbar = amean
        fold = lambda s, w, sum_w, n=None: \
            (w*(s-amean(s, w, sum_w))**2).sum() * sum_w / (sum_w**2 - (w**2).sum())
        unbiased_var = fold

    if methods['biased_sd']:
        # Biased standard deviation (requires biased_var)
        fold = lambda s, w, sum_of_w, n=None: \
            biased_var(s, w, sum_w)**0.5
        
    if methods['unbiased_sd']:
        # Unbiased standard deviation (requires unbiased_var)
        fold = lambda s, w, sum_w, n=None: \
            unbiased_var(s, w, sum_w)**0.5
        

    for s in spaces:

        # ----------------
        # Find the weights
        # ----------------
        if weights is not None:
            # Calculate the weights ...
            weights = calc_weights(s, collapse=collapse, wtype=wtype,
                                   infer_bounds=infer_bounds)
        
        else:
            # ... or use given weights
            s = s.reorder(weights)

        # ---------------------------------------
        # Reshape the space to (n, weights.shape)
        # ---------------------------------------
        s = s.n_by_shape(weights)
        
        # -------------------------------------------------------
        # Initialize the output collapsed space to with a 1d data
        # array. We'll reshape it later.
        # -------------------------------------------------------
        c = s.copy(data = numpy.empty(s.shape[0]))
        
        # Find a missing data indicator
        missing_data = None
        if hasattr(s, '_FillValue'):
            missing_data = s._FillValue
        elif hasattr(s, 'missing_value'): 
            missing_data = s.missing_value
            
        # ----------------------------------------------------
        # Collapse the data into a 1d space
        # ----------------------------------------------------
        if missing_data is not None:

            for n in xrange(c.size):

                # Mask the weights where there is missing data
                # (vweights means 'valid weights').
                vweights = numpy.ma.array(weights, mask = s[n,...].wne(missing_data))
                print vweights[...]

                # Find number of non zero weights
                n_non_zero = vweights.size - numpy.ma.count_masked(vweights)

                # If the data throughout all of the collapse
                # dimensions are missing data, then set the collapsed
                # value to missing data and continue
                if n_non_zero == 0:
                    c[n] = missing_data
                    continue

                c[n] = fold(s[n,...], vweights, float(vweights.sum()))

        else:

            sum_of_weights = float(weights.sum())
            for n in xrange(c.size):
                c[n] = fold(s[n,...], weights, sum_of_weights)
         
        # ---------------------------------------------------------------
        # Sort out the dimensions and coordinates of the collapsed
        # space
        # ---------------------------------------------------------------
        for dim in weights.dimensions:
            c.dimensions.remove(dim)        
            c.space_dimensions[dim] = 1
            
        for coord in weights.collapsed_coord:
            c.coord[coord] = weights.collapsed_coord[coord].copy()
            
        # Remove auxiliary coordinates which span a collapse dimension
        weights_dimensions = set(weights.dimensions)
        for aux in c.coord_list('aux'):
            if set(c.coord[aux].dimensions) & weights_dimensions:
                c.coord.pop(aux)
                
        # -----------------------------------------------------
        # Sort out cell_methods
        # -----------------------------------------------------
        for dim in weights.dimensions:
            if s.space_dimensions[dim] == 1: continue
            # cell_methods = 'dim: mean'
            
        # -----------------------------------------------------
        # Reshape the collapsed space, c
        # -----------------------------------------------------
        try:
            c = c.reshape(s.original_shape[:s.ndim-weights.ndim])
        except: pass
        
        spaces_out.append(c)

    #--- End: for s in spaces:

    return spaces_out
    
#--- End: def collapse     

def calc_weights(s, collapse=[], wtype={}, infer_bounds=False):

    '''Calculate weights for collapsing. See def collapse for help'''
    
    from math import sin, pi
  
    def equal(s, dim, coord, infer):
        '''Calculate equal weights'''
        return numpy.ones(s.space_dimensions[dim])
    #--- End: def equal

    def linear(s, dim, coord, infer):
        '''Calculate linear weights based on cell bounds'''
        bounds = find_bounds_(s, coord, infer=infer)
        return numpy.array([abs(y-x) for x,y in \
                                zip(bounds[:,0], bounds[:,1])])
    #--- End: def linear

    def latitude_area(s, dim, coord, infer):
        '''Calculate latitude-area weights based on cell bounds'''
        bounds = find_bounds_(s, coord, infer=infer)
        pi_over_180 = pi/180.0
        return numpy.array([abs(sin(y*pi_over_180)-sin(x*pi_over_180)) \
                                for x,y in zip(bounds[:,0], bounds[:,1])])
    #--- End: def latitude_area

    def weights_(input, s=None, dim=None, coord=None, infer=None):
        '''Find weights from 'input'.

        If 'input' is a 1d array (or cf variable) of weights, then
        just return the array unchanged.

        If 'input' is a string then execute the function of the same name.

        If 'input' is a function, then execute it.

        In the latter two cases, 'arg1', 'arg2', and 'infer' are the
        arguments to the function. '''

        try:
            # Input is a function
            return input(s, dim, coord, infer)
        except TypeError:
            try:
                input = eval(input)  
            except TypeError:
                # Input is an array
                n_array = numpy.array(input)
                if n_array.ndim > 1:
                    error('ERROR: can only supply 1d arrays')
                if n_array.size != s.space_dimensions[dim]:
                    error('ERROR: array must be same size as collapse dimension')
                return n_array 
            else:
                # Input was a string but is now a function
                return input(s, dim, coord, infer)
    #--- End: def weights_

    def find_bounds_(s, coord, infer=False):
        '''Given a coordinate variable ('coord') find its bounds,
        either by returning an existing bounds array or, if that
        doesn't exist, guessing the bounds as the midpoints of the
        coordinate values (with extrapolation at the end
        points). Failure to find a bounds array results in an error.'''

        try:
            c = s.coord[coord]
        except KeyError:
            error('ERROR: Can not find bounds for missing coordinate')

        bounds = None
        try:
            bounds = numpy.copy(c.bounds)
        except AttributeError: 
            # No bounds attribute
            if infer:
                try:
                    # Does the coordinate have at least 2 values? ...
                    c[1]
                except IndexError:        
                    # ... No
                    pass
                else:
                    # ... Yes, so we can try to infer bounds
                    try:
                        bounds = numpy.empty((c.size, 2))    
                        i = 0
                        bounds[i,0] = 1.5* c[i] - 0.5*c[i+1] 
                        bounds[i,1] = 0.5*(c[i] +     c[i+1])   
                        i = 1
                        while i < c.size - 1:
                            bounds[i,0] = 0.5*(c[i-1] + c[i])  
                            bounds[i,1] = 0.5*(c[i]   + c[i+1])
                            i += 1                
                        bounds[i,0] = 0.5*(c[i] +     c[i-1])
                        bounds[i,1] = 1.5* c[i] - 0.5*c[i-1] 
                    except TypeError:
                        # Can't do string valued coordinates
                        pass
                    
        # Sort out out-of-bounds latitude values
        if c.name() in ('latitude', 'grid_latitude'):           
            try:
                for i in (0,-1):
                    if not feq(abs(c[i]), 90.0): continue
                    if bounds[i,0] >  90.0: bounds[i,0] =  90.0
                    if bounds[i,0] < -90.0: bounds[i,0] = -90.0
                    if bounds[i,1] >  90.0: bounds[i,1] =  90.0
                    if bounds[i,1] < -90.0: bounds[i,1] = -90.0
            except TypeError: pass     

        if bounds is None:
            error('ERROR: Can not find bounds')

        return bounds            
    #--- End: def find_bounds_

    def collapsed_coord_(s, coord, bounds=None):
        '''Create a collapsed coordinate from a coordinate. The
        collapsed coordinate is a 1d, size coordinate with bounds and
        optionally a coordinate value. The new bounds give some
        measure of the extent of the uncollapsed coordinate values and
        the new coordinate value is the mean of the bounds, where
        possible (e.g. for string valued coordinates, the new
        coordinate value array = []).'''

        # Return None if no coordinate
        try:
            coord = s.coord[coord].copy()
        except KeyError:
            return None

        # Leave scalar coordinate unchanged
        if s.space_dimensions[coord.dimensions[0]] == 1:
            return coord

        # Delete empty bounds
        try:
            if coord.bounds.size == 0:
                delattr(coord, 'bounds')
        except AttributeError: pass
        
        # Set the collapsed bounds
        if bounds is None:
            
            try:
                # Numeric bounds       
                coord.bounds = numpy.array([coord.bounds.min(),
                                            coord.bounds.max()])
            except TypeError:      
                # String bounds                       
                coord.bounds = numpy.array([coord.bounds[ 0,0],
                                            coord.bounds[-1,1]])                
            except AttributeError:                              
                # No bounds attribute
                try:
                    # Coorindate array is non-empty?
                    coord[0]
                except IndexError:
                    try:
                        # Numeric coordinate values
                        coord.bounds = numpy.array([coord.min(), coord.max()])
                    except TypeError:
                        # String coordinate values
                        coord.bounds = numpy.array([coord[0], coord[-1]])
        else:
            # Use given array for bounds
            try:
                coord.bounds = numpy.array([bounds.min(), bounds.max()])
            except TypeError:
                # String values
                coord.bounds = numpy.array([bounds[0,0], bounds[-1,1]])

        try:
            coord.bounds = coord.bounds.reshape(1,2)
        except AttributeError: pass
        
        # Set the collapsed coordinate value 
        try:
            # Numeric bounds
            coord = coord.copy(data = numpy.array([coord.bounds.mean()]))
        except (TypeError, AttributeError):
            # String bounds or no bounds
            coord = coord.copy(data = numpy.array([]))

        return coord
            
    #--- End: def collapsed_coord_

    # --------------------------------------------------------------
    # Find the collapse dimensions and their coordinates from the
    # input strings (e.g. input string 'regio' might give
    # coord,dims='aux0',['dim3']).
    #
    # Link these, along with each dimension's collapse type in two
    # dictionaries. E.g. dim2coord['dim3']='aux0',
    # dim2wtype['dim3']='equal'
    # --------------------------------------------------------------
    dim2coord = {}
    dim2wtype = {}
    for string in collapse + wtype.keys():
        name, coord, dims = s.coordinate(string, grid_mapping=True, 
                                         info=True, noerror=True)
        if dims is None:
            # e.g. string = 'splatitude'
            error('ERROR: can not determine collapse dimension from '+string)    
        elif len(dims) > 1:
            # e.g. dims=['dim2','dim1']
            error('identified coordinate has multiple dimensions: '+coord)

        # Note: coord may equal None if there is no coordinate for the
        # identified dimension
        dim2coord[dims[0]] = coord
        dim2wtype[dim]     = wtype[string] if string in wtype else 'default'

        # Set the weight type for size 1 collapse dimensions to 'equal'
        if s.space_dimensions[dims[0]] == 1:
            dim2wtype[dim] = 'equal'

    # Create a list of all the collapse dimensions
    dimensions = dim2coord.keys()

    # Check that we're not trying to do something dangerous
    for dim in dim2wtype:
        if dim2wtype[dim] in ('latitude_area', 'linear', 'default'):
            if dim2coord[dim] not in s.coord_list():
                error('ERROR: Can\'t do '+dim2wtype[dim]+' weights for missing coords')
            if re.search('^string', s.coord[coord].dtype.name):
                error('ERROR: Can\'t do '+dim2wtype[dim]+' weights for string coords')

    # Set flags for the possible modification of polar row weights
    dim_lat, dim_lon = None, None
    for dim, coord, dim in dim2coord.iteritems():
        try:
            if s.coord[coord].name() in ('latitude', 'grid_latitude'):
                dim_lat = dim
            elif s.coord[coord].name() in ('longitude', 'grid_longitude'):
                dim_lon = dim
        except KeyError: pass

    # ---------------------------------------------------------------
    # Initialize the index of the partial weights list. Note that the
    # final partial weights list may have fewer elements than the
    # number of collapse dimensions. E.g. an appropiate three
    # dimensionsal cell_measures array will provide just one element
    # of the partial weights list.
    # ----------------------------------------------------------------
    partial_weights  = []
    collapsed_coord  = {}
    n = 0

    # ----------------------------------------------------------------
    # Use cell_measures array values as weights for their
    # dimensions. Such an array will only be used if all of its
    # dimensions are collapse dimensions.
    # ----------------------------------------------------------------
    try:
        # Is there a cell_measures dictionary? ...
        s.cell_measures
    except AttributeError:
        # ... No
        pass
    else:                  
        # ... Yes

        # Create a list of dimensions which have user-defined weight
        # types
        overrides = []
        for dim in dim2coord:
            if dim2wtype[dim] != 'default':
                overrides.append(dim)

        for cm in s.cell_measures:

            # Ignore cell_measures array if it contains a non-collapse
            # dimension.
            if not set(s.cell_measures[cm].dimensions) <= set(dim2coord):
                continue

            # Ignore cell_measures array if a specific weights type
            # has been set for at least one of te cell_measures
            # dimensions.
            if set(s.cell_measures[cm].dimensions) & set(overrides):
                continue

            # All of this cell_measures array's dimensions are
            # collapse dimensions, so add the array to the list of
            # partial weights.
            partial_weights.append(s.cell_measures[cm].copy())
            
            # Create a collapsed coordinate for each cell_measures
            # dimension and pop the coordinate from the dimensions
            # list
            for dim in s.cell_measures[cm].dimensions:
                coord = dim2coord[dim]
                collapsed_coord[coord] = collapsed_coord_(s, coord)
                dimensions.remove(dim)
            
            n += 1              
    #--- End: try: except: else:

    # --------------------------------------------------------------
    # Find the vertical component of weights for collapse dimensions
    # which have a formula_terms-derived auxiliary coordinate
    # --------------------------------------------------------------
    v_component_only = []
    for dim in dimensions[:]:
        
        # Only allow equal or linear weights (for now!)
        if dim2wtype[dim] == 'equal':
            # Equal weights required, so leave this dimension to the
            # 1d case below
            continue

        if dim2wtype[dim] not in ('linear', 'default'):
            error('ERROR: Can only have equal, linear or default weights for formula_terms')

        # Try to create a new auxiliary coordinate from formula_terms,
        # including vertical bounds (vbounds) if possible
        coord = dim2coord[dim]
        aux = s.dimensional_vert_coord(coord, infer=infer_bounds, noerror=True)
        
        # No formula_terms-derived auxiliary coordinate, so leave this
        # dimension to the 1d case
        if aux is None: continue                

        # Find the linear vertical component of weights
        try:
            # Does aux have vbounds? ...
            aux.vbounds
        except AttributeError:
            # ... No
            error('ERROR: Can\'t find any bounds for '+aux.name())
        else:
            # ... Yes, so lets calculate the vertical weights
            weights = numpy.empty(aux.shape)
            for k in xrange(weights.shape[0]):
                weights[k,...] = abs(aux.vbounds[k,...,0] - aux.vbounds[k,...,1])
                
        # Add the vertical-component-of-weights array to the list of
        # partial weights
        partial_weights.append(new_variable(data       = weights,
                                            properties = {'dimensions': aux.dimensions}))
        
        # Create a coordinate for the collapsed vertical dimension
        # (collapsed coordinates for any other dimensions of the
        # weights array, if required, are created elsewhere)
        try:
            collapsed_coord[coord] = collapsed_coord_(s, coord, bounds=s.coord[coord].bounds)
        except (KeyError, AttributeError):
            collapsed_coord[coord] = None

        if aux.ndim > 1:
            v_component_only.append(n)
            
        dimensions.remove(dim)
        n += 1
    #--- End: for dim in dimensions[:]:
        
    # ------------------------------------------------------------
    # Find 1d weights for the remaining collapse dimensions
    # ------------------------------------------------------------

    # Order the list of the remaining collapse dimensions so that they
    # appear in the same order as the same dimensions in the
    # space. [This is a quick convienience that may reduce the amount
    # of reordering that has to be done on the space (or should that
    # be weights?)  data array (but reordering numpy arrays might be
    # very quick, anyway?  But it might involve a copy ...).]
    indices    = [s.dimensions.index(dim) for dim in dimensions]
    reorder    = [indices.index(n) for n in sorted(indices)]
    dimensions = [dimensions[n]    for n in reorder]
            
    for dim in dimensions[:]:

        # Set the coordinate and name of the collapse dimension
        # (e.g. coord,name='dim1','latitude')
        coord = dim2coord[dim]
        try:
            name = s.coord[coord].name()
        except KeyError:
            name = None

        # --------------------------------------------------------
        # CASE 0: Specifically requested weights 
        # --------------------------------------------------------
        if dim2wtype[dim] != 'default':
 
            weights = weights_(dim2wtype[dim], s, dim, coord, infer=infer_bounds)
            
        # --------------------------------------------------------
        # CASE 1: Latitude weights (add new names to list)
        # --------------------------------------------------------
        elif name in ('latitude', 'grid_latitude'):
            
            weights = weights_(latitude_area, s, dim, coord, infer=infer_bounds)

        # --------------------------------------------------------
        # CASE 2: General linear weights (add new names to list)
        # --------------------------------------------------------
        elif name in ('longitude', 'grid_longitude',
                      'depth', 'air_pressure', 'height'
                      'time'):

            weights = weights_(linear, s, dim, coord, infer=infer_bounds)
            
        # --------------------------------------------------------
        # CASE 3: General equal weights (add new names to list)
        # --------------------------------------------------------
        elif name in ():

            weights = weights_(equal, s, dim, coord, infer=infer_bounds)
            
        # ----------------------------------------------------------
        # CASE 4: Linear weights for projection coordinates (add new
        #         names to list)
        # ----------------------------------------------------------
        elif name in ('projection_x_coordinate, projection_y_coordinate'):
            
            if set(s.grid_mapping) & set('albers_conical_equal_area',      # ? 
                                         'lambert_cylindrical_equal_area', # ?
                                         'lambert_azimuthal_equal_area'):  # ?
                weights = weights_(linear, s, dim, coord, infer=infer_bounds)
             
        # --------------------------------------------------------
        # CASE N: Add code for new types here
        # --------------------------------------------------------

        # Fail if weights have not been calculated
        else:
            error('ERROR: Can\'t calc weights for '+name)
                
        # Add the 1d weights array to the list of partial weights
        partial_weights.append(new_variable(data       = weights,
                                            properties = {'dimensions': [dim]}))
        
        # Create a coordinate for the collapsed dimension
        collapsed_coord[coord] = collapsed_coord_(s, coord, bounds=bounds)
    
        dimensions.remove(dim)
        n += 1
    #--- End: for dim in dimensions:
        
    # Adjust latitude-longitude weights for polar rows (i.e. rows with
    # grid box centres at +-90 degrees)
    if set((dim_lon, dim_lat)) <= set(dim2coord):
        for i in xrange(len(partial_weights)):
            if dim_lat not in partial_weights[i].dimensions: continue
            if i in v_component_only: continue
            for j in (0,-1):
                if feq(abs(s.coord[dim_lat][j]), 90.0):
                    partial_weights[i] = partial_weights[i].reorder(dim_lat, 
                                                                    same_dims = True, 
                                                                    left      = True)
                    partial_weights[i][j,...] /= s.coord[dim_lon].size

    # ----------------------------------------------------------------
    # If there are multidimensionsal partial weights arrays which just
    # contain just a vertical component (given by the v_component_only
    # list of indices), then subsume the weights from other relevant
    # dimensions into the multidimensionsal array values.
    # ----------------------------------------------------------------
    use_partial_weight  = range(len(partial_weights))
    all_partial_weights = use_partial_weight[:]
    for j in v_component_only:

        for i in all_partial_weights:
            
            if i == j: continue

            if set(partial_weights[i].dimensions) < set(partial_weights[j].dimensions):
                
                # The dimensions of partial_weights[i] are a subset
                # (not equal) to those of vertical component weights
                # (partial_weights[j]), so introduce the non-vertical
                # weights from [i] into [j]

                # Reorder and reshape to shape=(n,partial_weights[i].shape)
                partial_weights[j] = partial_weights[j].reorder(partial_weights[i], 
                                                                same_dims = True)
                partial_weights[j] = partial_weights[j].n_by_shape(partial_weights[i])
                
                # Multiply the vertical weights by those in the other
                # dimensions
                for n in xrange(partial_weights[j].shape[0]):
                    partial_weights[j][n,...] *= partial_weights[i]

                # Restore the shape
                partial_weights[j] = partial_weights[j].reshape(partial_weights[j].original_shape)

                # Flag not to use the subsumed partial weights array ([i])
                use_partial_weight.pop(i)
     
    # -------------------------------------------------------------------
    # Create the final, multidimensional weights array from the outer
    # product of of the partial weights arrays stored in
    # partial_weights[:]
    # -------------------------------------------------------------------
    weights    = partial_weights[use_partial_weight[0]]
    shape      = partial_weights[use_partial_weight[0]].shape
    dimensions = partial_weights[use_partial_weight[0]].dimensions
    use_partial_weight.pop(0)
    for n in use_partial_weight:
        # Note: Order of outer product matters for the later reshape
        weights     = numpy.outer(weights, partial_weights[n])
        shape      += partial_weights[n].shape
        dimensions += partial_weights[n].dimensions
    
    # Turn weights into a cf variable. Reshape it so that it has the
    # same rank as the collapse dimensions and reorder it so that its
    # dimensions appear in the same order as the same dimensions in
    # the space
    weights = new_variable(data       = weights.reshape(shape), 
                           properties = {'dimensions': dimensions,
                                         'long_name' : 'collapse_weights'})
    weights = weights.reorder(s, same_dims=True)

    # Give weights dictionaries of the full and collapsed coordinates
    weights.collapsed_coord = {}
    for coord in collapsed_coord:
        if collapsed_coord[coord] is not None:
            weights.collapsed_coord[coord] = collapsed_coord[coord]

    weights.coord = {}
    for coord in dim2coord.values():
        try:
            weights.coord[coord] = s.coord[coord].copy()
        except KeyError: pass

    # Return the weights
    return weights

#--- End: def calc_weights     
