import copy
import numpy
import re
import netCDF4
import textwrap
from .utils import RTOL, ATOL, equals, Comparison, CfList, CfDict
from .units import Units

def _parse_index(x, index):
    '''

Private function used by Variable and its subclasses in their
__getitem__ methods.

Expand the index input to __getitem__ so that it has one slice per
dimension of the data of `s`, a variable.
    
    '''
    # Fix index, handling ellipsis and incomplete slices (adapted from
    # numpy)
    if not isinstance(index, tuple):
        index = (index,)
    fixed = []
    one_dim_at_a_time = 0
    length, dims = len(index), len(x.shape)
    for slice_ in index:

        if slice_ is Ellipsis:
            fixed.extend([slice(None)] * (dims-length+1))
            length = len(fixed)
        elif isinstance(slice_, (int, long)):
            if slice_ == -1:                    
                fixed.append(slice(slice_, None, 1))
            else:
                fixed.append(slice(slice_, slice_+1, 1))
        elif isinstance(slice_, slice):
            fixed.append(slice_)
        else:
            if isinstance(slice_, tuple):
                slice_ = list(slice_)
                
            one_dim_at_a_time += 1
            
            # Replace a boolean array with a sequence of integers
#            if getattr(getattr(slice_, 'dtype', None), 'kind', None) == 'b':
            try: 
                if slice_.dtype.kind == 'b':
                    slice_ = numpy.where(slice_)[0]           
            except AttributeError:
                pass

            # Replace a sequence of positive integers with a slice, if
            # possible (could onvert -ve ints to positive ones, but
            # the permanent overhead is probably not worth it???)
            if len(slice_) == 1:
                # The sequence contains just one integer - always
                # replace with a slice
                if slice_[0] == -1:                    
                    slice_ = slice(-1, None, 1)
                else:
                    slice_ = slice(slice_[0], slice_[0]+1, 1)
                one_dim_at_a_time -= 1                    
            else:
                # The sequence contains at least two integers - see if
                # we can replace it with a slice
                start = slice_[0]
                stop  = slice_[-1]
                if start >= 0 <= stop:
                    step  = slice_[1] - slice_[0]
                    stop += step
                    len_slice = (stop - start)/step
                    if stop < 0:
                        stop = -1
                    if (len(slice_) == len_slice and 
                        (numpy.arange(start, stop, step) == slice_).all()):
                        if stop == -1:
                            stop = None
                        slice_ = slice(start, stop, step)
                        one_dim_at_a_time -= 1
                        
            fixed.append(slice_)
        #--- End: if
    #--- End: for

    index = tuple(fixed)

    # Fill in missing dimensions with slice(None)
    if len(index) < dims:
        index += (slice(None),) * (dims-len(index))
            
    # If fewer than 2 dimensions are sliced by a sequence of integers
    # then we don't need to apply the slice one dimension at a time
    if one_dim_at_a_time < 2:
        one_dim_at_a_time = 0

    return index, one_dim_at_a_time
#--- End: def


# ====================================================================
#
# Variable objects
#
# ====================================================================

class Variable(object):
    '''
    
Base class for a space and its components.

Refer to the :ref:`variable data storage, access and manipulation
<variable_data_storage_access_and_manipulation>` in the top level
:mod:`cf` documentation for more details.

:keyword public_attributes: *Optional*. The new variable is given the
                            public attributes named and valued from
                            the key, value pairs of this
                            dictionary. The values are deep copied.

:type public_attributes: dict

:keyword private_attributes: *Optional*. The new variable is given the
                             private attributes named and valued from
                             the key, value pairs of this
                             dictionary. The values are deep copied.

:type private_attributes: dict

**Read only attributes**

The following attributes are read only on the variable and may not be
set.

============== ========================================  
Attribute      Description                             
============== ========================================
:attr:`array`  A numpy array deep copy of the data.    
:attr:`dtype`  Describes the format of the             
               elements in the data array (refer to          
               :obj:`numpy.ndarray.dtype`).                 
:attr:`ndim`   The data array's number of dimensions   
               (refer to :obj:`numpy.ndarray.ndim`).        
:attr:`shape`  The shape of the data array             
               (refer to :obj:`numpy.ndarray.shape`).       
:attr:`size`   The number of elements in the data array
               (refer to :obj:`numpy.ndarray.size`).        
:attr:`varray` A numpy view of the data.              
============== ========================================


**Reserved attributes:**

The following attributes have special definitions and it is
recommended to set them accordingly.

============== ======================================================
Attribute      Description                         
============== ======================================================
:attr:`_data`  The variable's :class:`~cf.Data` object.
:attr:`_Units` The variable's :class:`~cf.Units` object. Setting this
               changes the variable's `calendar` and `units` public
               attributes, and vice versa.
============== ======================================================


**Special private attributes:**

The following private attributes have special definitions and it is
recommended to set them accordingly.

============== ======================================================
Attribute      Description                         
============== ======================================================
:attr:`file`   DCH
:attr:`ncvar`  The variable's netCDF variable name.
============== ======================================================


**Reserved public attributes:**

The following public attributes have identical descriptions as in the
CF convention and may be set directly on the instance as well as with
the :meth:`setpub` or :meth:`pub` methods and retrieved directly from
the instance as well as with the :meth:`getpub` or :meth:`pub`
methods.

===================== ========================
Attribute             Description             
===================== ========================
:attr:`calendar`      As in the CF convention.
:attr:`_FillValue`    As in the CF convention.
:attr:`long_name`     As in the CF convention.
:attr:`missing_value` As in the CF convention.
:attr:`standard_name` As in the CF convention.
:attr:`units`         As in the CF convention.
===================== ========================


**Methods:**

A variable has the following methods.

================================ ================================================
Method                           Description
================================ ================================================
:meth:`~cf.Variable.copy`        Create a deep copy of the variable.

:meth:`~cf.Variable.dump`        Return a string containing a full description of
                                 the variable.

:meth:`~cf.Variable.equals`      Determine whether two instances are congruent
                                 with each other.

:meth:`~cf.Variable.match`       Determine whether a variable matches phenomena
                                 criteria.

:meth:`~cf.Variable.first_datum` Return the first element of the data without
                                 replacing a file pointer with a numpy array.

:meth:`~cf.Variable.last_datum`  Return the last element of the data without
                                 replacing a file pointer with a numpy array.

:meth:`~cf.Variable.name`        Return the variable's name.

:meth:`~cf.Variable.type`        The type of the data stored in the variable's
                                 :class:`~cf.Data` object.

:meth:`~cf.Variable.pubattrs`    Return a dictionary of the variable's public
                                 attributes.

:meth:`~cf.Variable.privattrs`   Return a dictionary of the variable's private
                                 attributes.

:meth:`~cf.Variable.istime`
:meth:`~cf.Variable.ematch`
:meth:`~cf.Variable.extract`
:meth:`~cf.Variable.eextract`
:meth:`~cf.Variable.priv`        Set or get a private attribute.        
:meth:`~cf.Variable.setpriv`     Set a private attribute.            
:meth:`~cf.Variable.getpriv`     Get a private attribute.            
:meth:`~cf.Variable.delpriv`     Delete a private attribute.         
:meth:`~cf.Variable.pub`         Set or get a public attribute.
:meth:`~cf.Variable.setpub`      Set a public attribute.
:meth:`~cf.Variable.getpub`      Get a public attribute.
:meth:`~cf.Variable.delpub`      Delete a public attribute.         
================================ ================================================


**Methods and attributes defined here:**
'''

    _private_attrs = ('_private_attrs',
                      '_data_insert'
                      '_concatenate_data',
                      '_getter',
                      '_setter',
                      '_deleter',
                      '_match',
                      'pubattrs',
                      'privattrs',
                      'first_datum',
                      'last_datum',
                      'type',
                      'copy',
                      'istime',
                      'dump',
                      'equals',
                      'match',
                      'ematch',
                      'extract',
                      'eextract',
                      'setpriv',
                      'getpriv',
                      'delpriv',
                      'pub',
                      'setpub',
                      'getpub',
                      'delpub',
                      'name',
                      )
    
    def __init__(self,
                 public_attributes={},
                 private_attributes={}):
        '''

Set the public and private attributes from the arguments.
'''
        self._private = {}    

        for attr, value in public_attributes.iteritems():
            self.setpub(attr, copy.deepcopy(value))        

        for attr, value in private_attributes.iteritems():
            self.setpriv(attr, copy.deepcopy(value))        
    #--- End: def

    def __setattr__(self, attr, value):
        '''
x.__setattr__(attr, *default) <==> x.attr=value
'''
        if attr in self._private_attrs:
            raise AttributeError, "Can't set reserved attribute '%s'" % attr

        super(Variable, self).__setattr__(attr, value) 
    #--- End: def

    def __delattr__(self, attr):
        '''
x.__delattr__(attr) <==> del x.attr
'''
        if attr in self._private_attrs:
            raise AttributeError, "Can't delete reserved attribute '%s'" % attr

        super(Variable, self).__delattr__(attr) 
    #--- End: def

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
'''

        if index == slice(None):
            return self

        try:
            if (index.start not in (0, -1) or
                index.stop != 1 or
                (index.step is not None or index.step < 1)):
                raise IndexError, "Bad '%s' index: %s" % \
                    (self.__class__.__name__, index)
        except AttributeError:
            if index not in (0, -1):
                raise IndexError, "%s index out of range: %s" % \
                    (self.__class__.__name__, index)
                
        return self
    #--- End: def

    def __setitem__(self, index, value):
        '''
x.__setitem__(index, value) <==> x[index]=value
'''
        pass # for now! DCH
    #--- End: def

    def __len__(self):
        '''
x.__len__() <==> len(x)
'''
        return 1
    #--- End: def

    def __deepcopy__(self, memo):
        '''

Used if copy.deepcopy is called on the variable.

.. seealso:: :meth:`~cf.Variable.copy`.
''' 
        return self.copy()
    #--- End: def

    def __add__(self, other):
        '''
x.__add__(y) <==> x+y
'''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new = self.copy(data=False)
        new._data._list[0] = x.__add__(y)
        return new
    #--- End: def

    def __sub__(self, other):
        '''
x.__sub__(y) <==> x-y
'''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new = self.copy(data=False)
        new._data._list[0] = x.__sub__(y)
        return new
    #--- End: def

    def __mul__(self, other):
        '''
x.__mul__(y) <==> x*y
'''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__mul__(y)
        return new
    #--- End: def

    def __div__(self, other):
        '''
x.__div__(y) <==> x/y
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__div__(y)
        return new
    #--- End: def

    def __truediv__(self, other):
        '''
x.__truediv__(y) <==> x/y
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new = self.copy(data=False)
        new._data._list[0] = x.__truediv__(y)
        return new
    #--- End: def

    def __floordiv__(self, other):
        '''
x.__div__(y) <==> x//y
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__floordiv__(y)
        return new
    #--- End: def
        
    def __mod__(self, other):
        '''
x.__mod__(y) <==> x%y      
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__mod__(y)
        return new
    #--- End: def

    def __pow__(self, other, modulo=None):
        '''
x.__pow__(y) <==> x**y
    '''
        if modulo is not None:
            raise RuntimeError('3-argument power not supported')
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__pow__(y, modulo)
        return new
    #--- End: def
    
    def __and__(self, other):
        '''
x.__and__(y) <==> x&y
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__and__(y)
        return new
    #--- End: def
    
    def __xor__(self, other):
        '''
x.__xor__(y) <==> x^y
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__xor__(y)
        return new
    #--- End: def
    
    def __or__(self, other):
        '''
x.__or__(y) <==> x|y
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__or__(y)
        return new
    #--- End: def

    def __iadd__(self, other):
        '''
x.__iadd__(y) <==> x+=y
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        x.__iadd__(y)
        return self
    #--- End: def

    def __isub__(self, other):
        '''
x.__isub__(y) <==> x-=y
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        x.__isub__(y)
        return self
    #--- End: def

    def __imul__(self, other):
        '''
x.__imul__(y) <==> x*=y
'''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        x.__imul__(y)
        return self
    #--- End: def

    def __idiv__(self, other):
        '''
x.__idiv__(y) <==> x/=y
'''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        x.__idiv__(y)
        return self
    #--- End: def

    def __itruediv__(self, other):
        '''
x.__itruediv__(y) <==> x/=y
'''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        x.__itruediv__(y)
        return self
     #--- End: def

    def __ifloordiv__(self, other):
        '''
x.__ifloordiv__(y) <==> x//=y
'''        
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        x.__ifloordiv__(y)
        return self
    #--- End: def
        
    def __imod__(self, other):
        '''
x.__imod__(y) <==> x%=y
'''        
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        x.__imod__(y)
        return self
    #--- End: def

    def __ipow__(self, other, modulo=None):
        '''
x.__ipow__(y) <==> x**=y
'''        
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        x.__ipow__(y)
        return self
    #--- End: def
    
    def __iand__(self, other):
        '''
x.__iand__(y) <==> x&=y
'''     
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        x.__iand__(y)
        return self        
    #--- End: def
    
    def __ixor__(self, other):
        '''
x.__ixor__(y) <==> x^=y
'''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        x.__ixor__(y)
        return self
    #--- End: def
    
    def __ior__(self, other):
        '''
x.__ior__(y) <==> x|=y
'''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        x.__ior__(y)
        return self
    #--- End: def

    def __radd__(self, other):
        '''
x.__radd__(y) <==> y+x
'''       
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__radd__(y)
        return new
    #--- End: def

    def __rsub__(self, other):
        '''
x.__rsub__(y) <==> y-x  
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__rsub__(y)
        return new
    #--- End: def

    def __rmul__(self, other):
        '''
x.__rmul__(y) <==> y*x
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__rmul__(y)
        return new
    #--- End: def

    def __rdiv__(self, other):
        '''
x.__rdiv__(y) <==> y/x
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__rdiv__(y)
        return new
    #--- End: def

    def __rtruediv__(self, other):
        '''
x.__rtruediv__(y) <==> y/x
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new          = self.copy(data=False)
        new._data._list[0] = x.__rtruediv__(y)
        return new
    #--- End: def

    def __rfloordiv__(self, other):
        '''
x.__rfloordiv__(y) <==> y//x
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        new = self.copy(data=False)
        new._data._list[0] = x.__rfloordiv__(y)
        return new
    #--- End: def
        
    def __rmod__(self, other):
        '''
x.__rmod__(y) <==> y%x
    '''
        x = self.varray
        try:
            y = other.varray
        except AttributeError:
            y = other
        x   = self.varray
        new = self.copy(data=False)
        new._data._list[0] = x.__rmod__(y)
        return new
    #--- End: def

    def __neg_(self):
        '''
x.__neg__() <==> -x
    '''
        return self * -1
    #--- End: def

    def __pos__(self):
        '''
x.__pos__() <==> +x
    '''
        return self.copy()
    #--- End: def

    def __abs__(self):
         '''
x.__abs__() <==> abs(x)
        '''
         x   = self.varray
         new = self.copy(data=False)    
         new._data._list[0] = numpy.absolute(x)
         return new
    #--- End: def

    def __invert__(self):
         '''
x.__abs__() <==> ~x
    '''
         x   = self.varray
         new = self.copy(data=False)
         new._data._list[0] = ~x
         return new
    #--- End: def
    
    def __eq__(self, other):
        '''
x.__eq__(y) <==> x==y
    '''            
        return self.__data_equals__(other)
    #--- End: def

    def __data_equals__(self, other, rtol=None, atol=None):
        '''

As :meth:`~cf.Variable.__eq__` but allowing the numeric equality
tolerances to be set explicitly.

    '''
        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()
                
        try:
            other._data

        except AttributeError:
            # -------------------------------
            # `other` is a scalar or an array
            # -------------------------------
            try:
                # Numeric values
                return numpy.less_equal(numpy.absolute(self.varray-other),
                                        atol + rtol*numpy.absolute(other))
            except TypeError:
                # String values
                return self.varray == other  
        
        else: 
            # ---------------------
            # `other` is a Variable
            # ---------------------
            # Check for file pointers being the same or for one array
            # being a view of the other.
            if self._data is other._data:
                a = numpy.empty(self.shape, dtype=bool)
                a.fill(True)
                return a

            try:
                self._data.base
                other._data.base
            except AttributeError:
                pass
            else:
                if ((self._data.base is not None and self._data.base is other._data) or
                    (other._data.base is not None and other._data.base is self._data[0])):
                    a = numpy.empty(self.shape, dtype=bool)
                    a.fill(True)
                    return a

            x = self.varray
            y = other.varray
            try:
                # Numeric values
                return numpy.less_equal(numpy.absolute(x - y),
                                        atol + rtol*numpy.absolute(y))
            except TypeError:
                # String values
                return x == y
    #--- End: def

    def __ne__(self, other): 
        '''
x.__ne__(y) <==> x!=y
'''
        return ~(self == other)
    #--- End: def

    def __le__(self, other):
        '''
x.__le__(y) <==> x<=y
'''
        try:
            return other.varray >= self.varray
        except AttributeError:
            return other >= self.varray
    #--- End: def

    def __lt__(self, other):
        '''
x.__lt__(y) <==> x<y
'''
        try:
            return other.varray > self.varray
        except AttributeError:
            return other > self.varray
    #--- End: def

    def __ge__(self, other):
        '''
x.__ge__(y) <==> x>=y
'''
        return ~(self < other)
    #--- End: def

    def __gt__(self, other):
        '''
x.__gt__(y) <==> x>y
'''
        return ~(self <= other)
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''   
        class_name = self.__class__.__name__
        try:
            name = self.name(default='', long=True, ncvar=True)
        except TypeError:
            # DCH This is so, very bad. The whol cfatts things need rethinking
            name = self.name
        try:
            shape = str(self.shape).replace(',)', ')', 1)
            return '<CF %s: %s%s>' % (self.__class__.__name__, name, shape)
        except AttributeError:
            return '<CF %s: %s>' % (self.__class__.__name__, name)
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''   
        return self.__repr__()
    #--- End: def

    def _data_insert(self, indices, other,
                     self_dimensions, other_dimensions, 
                     match_for_dim1):
        '''
'''
        # ------------------------------------------------------------
        # Create a mapping between order names in other and those in
        # self. E.g. if self and other have dimension standard names
        # repectively:
        #
        #     ['time','latitude','longitude',-1]
        #     ['longitude','time',-1]
        #     
        # and
        #     
        #     self._data.order  = ['dim0','dim1','dim2','bounds']
        #     other._data.order = ['dim8','dim6','bounds']
        #     
        # then
        #    
        #     order_map['dim8']   = 'dim2'
        #     order_map['dim6']   = 'dim0'
        #     order_map['bounds'] = 'bounds'
        # ------------------------------------------------------------
#        print '\ndch sn=', self.pub('standard_name'), self.shape, other.shape
#        new_order  = []
#        for dim1 in other_dimensions:
#            try:
#                new_order.append(self_dimensions.index(match_for_dim1[dim1]))
#            except ValueError:
#                pass     
        #--- End: for
#        print 'dch self_dimensions =', self_dimensions, self._data.order 
#        print 'dch other_dimensions=', other_dimensions, other._data.order
#        print 'dch match_for_dim1=',match_for_dim1
#        print 'dch new_order=', new_order    
        order_map = {}

        for x in other._data.order:
            order_map[x] = match_for_dim1[x]

#        for i, n in enumerate(new_order):
#            try:
#                order_map[other._data.order[i]] = self._data.order[n]
#            except IndexError:
#                pass
            
        adim  = self._data.adim
        shape = list(self._data.shape)

        for i, n in enumerate(indices):

            # Insert the data and its metadata
            self._data.insert(n, other._data[i])

            # Change the dimension names of the newly inserted data to
            # conform with 'self._data'
            self._data.elements[n].order = [order_map[d]
                                            for d in self._data.elements[n].order]

            # update self.shape
            try:
                j = self._data.elements[n].order.index(self._data.order[adim])
            except ValueError: 
                shape[adim] += 1
            else:
                shape[adim] += self._data.elements[n].shape[j]
        #--- End: for

        self._data.shape = tuple(shape)
    #--- End: def

    def _concatenate_data(self):
        '''

'''
        try:
            data = self._data
        except AttributeError:
            pass
        else:
            data[:] = data.concatenate()
    #--- End:

    def _getter(self, attr):
        '''

Private function for getting attributes from the _private dictionary.
'''
        try:
            return self._private[attr]
        except (KeyError, AttributeError):
            raise AttributeError, "Can't get '%s' attribute '%s'" % \
                (self.__class__.__name__, attr)       
    #--- End: def

    def _setter(self, attr, value):
        '''

Private function for setting attributes in the _private dictionary.
'''
        self._private[attr] = value
    #--- End: def

    def _deleter(self, attr):
        '''

Private function for deleting attributes from the _private dictionary.
'''
        try:
            del self._private[attr]
        except KeyError:
            raise AttributeError, "Can't delete '%s' attribute '%s'" % \
                (self.__class__.__name__, attr)       
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: _data
    # ---------------------------------------------------------------- 
    @property
    def _data(self):        
        '''

The variable's Data object.
'''
        return self._getter('_data')
    #--- End: def
    @_data.setter
    def _data(self, value):
        '''

Set `_data` in the `_private` dictionary, overwriting `_Units` if the
new data value has it set.
'''
        self._setter('_data', value)
        try:
            if self._data._Units:
                try:
                    del self._private['_Units']
                except KeyError:
                    pass
        except AttributeError:
            pass
    #--- End: def

    @_data.deleter
    def _data(self):        self._deleter('_data')

    # ----------------------------------------------------------------
    # Property attribute: _Units
    # ----------------------------------------------------------------
    @property
    def _Units(self):
        '''

The variable's `_Units` object.
'''
        try:
            return self._data._Units
        except AttributeError:
            try:
                return self._private['_Units']
            except (KeyError, AttributeError):
                pass
        #--- End: try

        # Still here?
        raise AttributeError, "Can't get '%s' attribute '_Units'" % \
            self.__class__.__name__
    #--- End: def

    @_Units.setter
    def _Units(self, value):
        if not isinstance(value, Units):
            raise TypeError, "Can only set '_Units' attribute to a Units object"

        try:
            self._data._Units = value
        except AttributeError:
            self._private['_Units'] = value
     #--- End: def

    @_Units.deleter
    def _Units(self):
        try:
            del self._data._Units
            return None            
        except AttributeError:
            try:
                del self._private['_Units']
                return None            
            except AttributeError:               
                pass
        #--- End: try

        # Still here?
        raise AttributeError, "Can't delete '%s' attribute '_Units'" % \
            self.__class__.__name__            
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: _FillValue
    # ----------------------------------------------------------------
    @property
    def _FillValue(self):
        '''

The variable's CF `_FillValue` public attribute.

Assigning a value to this attribute also assigns the same value to the
`missing_value` public attribute, if the latter has previously been
defined.
'''
        try:
            return self._data._FillValue
        except AttributeError:
            try:
                return self._private['_FillValue']
            except (KeyError, AttributeError):
                pass
        #--- End: try

        # Still here?
        raise AttributeError, "Can't get '%s' attribute '_FillValue'" % \
            self.__class__.__name__
    #--- End: def

    @_FillValue.setter
    def _FillValue(self, value):
        try:
            self._data._FillValue = value
        except AttributeError:
            self._private['_FillValue'] = value
     #--- End: def

    @_FillValue.deleter
    def _FillValue(self):
        try:
            del self._data._FillValue
            return None            
        except AttributeError:
            try:
                del self._private['_FillValue']
                return None            
            except AttributeError:               
                pass
        #--- End: try

        # Still here?
        raise AttributeError, "Can't delete '%s' attribute '_FillValue'" % \
            self.__class__.__name__            
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: units
    # ----------------------------------------------------------------
    # DCH possible inconsistency when setting self._Units.units
    @property
    def units(self):
        '''

The variable's CF `units` public attribute.

x.units=value <==> x.setpub('units', value)
x.units       <==> x.getpub('units')
del x.units   <==> x.delpub('units')
x.units=value <==> x._Units.units=value
x.units       <==> x._Units.units
del x.units   <==> del x._Units.units
'''
        try:
            units = self._data._Units.units
            if units:
                return units
        except AttributeError:
            try:
                units = self._private['_Units'].units
                if units:
                    return units                
            except (KeyError, AttributeError):
                pass
        #--- End: try

        # Still here?
        raise AttributeError, "Can't get '%s' attribute 'units'" % \
            self.__class__.__name__
    #--- End: def

    @units.setter
    def units(self, value):
        try:
            data = self._data
        except AttributeError:
            try:
                self._private['_Units'].units = value
            except KeyError:
                self._private['_Units'] = Units(units=value)        
        else:
            try:
                data._Units.units = value
            except AttributeError:
                data._Units = Units(units=value)
     #--- End: def

    @units.deleter
    def units(self):
        try:
            del self._data._Units.units
            return None
        except AttributeError:
            try:
                del self._private['_Units'].units
                return None            
            except (KeyError, AttributeError):
                pass
        #--- End: try
#        try:
#            _Units = self._data._Units
#        except AttributeError:
#            try:
#                del self._private['_Units'].units
#                return None            
#            except (KeyError, AttributeError):
#                pass
#        else:
#            try:
#                del _Units.units
#                return None
#            except AttributeError:
#                pass
#        #--- End: try

        # Still here?
        raise AttributeError, "Can't delete '%s' attribute 'units'" % \
            self.__class__.__name__            
    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: calendar
    # ----------------------------------------------------------------
    @property
    def calendar(self):
        '''

The variable's CF `calendar` public attribute.

x.calendar=value <==> x.setpub('calendar', value)
x.calendar       <==> x.getpub('calendar')
del x.calendar   <==> x.delpub('calendar')
x.calendar=value <==> x._Units.calendar=value
x.calendar       <==> x._Units.calendar
del x.calendar   <==> del x._Units.calendar
'''
        try:
            return self._data._Units.calendar
        except AttributeError:
            try:
                return self._private['_Units'].calendar
            except (KeyError, AttributeError):
                pass
        #--- End: try

        raise AttributeError, "Can't get '%s' attribute 'calendar'" % \
            self.__class__.__name__
    #--- End: def

    @calendar.setter
    def calendar(self, value):       
        try:
            data = self._data
        except AttributeError:
            try:
                self._private['_Units'].calendar = value
            except (KeyError, AttributeError):
                self._private['_Units'] = Units(calendar=value)
        else:
            try:
                data._Units.calendar = value
            except AttributeError:
                data._Units = Units(calendar=value)
     #--- End: def

    @calendar.deleter
    def calendar(self):
        try:
            del self._data._Units.calendar
            return None
        except AttributeError:
            try:
                del self._private['_Units'].calendar
                return None            
            except (KeyError, AttributeError):
                pass
        #--- End: try
#       try:
#            _Units = self._data._Units
#        except AttributeError:
#            try:
#                del self._private['_Units'].calendar
#                return None            
#            except (KeyError, AttributeError):
#                pass
#        else:
#            try:
#                del _Units.calendar
#                return None    
#            except AttributeError:
#                pass
        #--- End: try

        # Still here?
        raise AttributeError, "Can't delete '%s' attribute 'calendar'" % \
            self.__class__.__name__            
    #--- End: def


    # ----------------------------------------------------------------
    # Property attribute: standard_name
    # ----------------------------------------------------------------
    @property
    def standard_name(self):
        '''

The variable's CF `standard_name` public attribute. 
'''
        return self.getpub('standard_name')
    #--- End: def
    @standard_name.setter
    def standard_name(self, value): self.setpub('standard_name', value)
    @standard_name.deleter
    def standard_name(self):        self.delpub('standard_name')

    # ----------------------------------------------------------------
    # Property attribute: long_name
    # ----------------------------------------------------------------
    @property
    def long_name(self):
        '''

The variable's CF `long_name` public attribute. 
'''
        return self.getpub('long_name')
    #--- End: def
    @long_name.setter
    def long_name(self, value): self.setpub('long_name', value)
    @long_name.deleter
    def lon_name(self):         self.delpub('long_name')

    # ----------------------------------------------------------------
    # Property attribute: missing_value
    # ----------------------------------------------------------------
    @property
    def missing_value(self):
        '''

The variable's CF `missing_value` public attribute.

Assigning a value to this attribute also assigns the same value to the
`_FillValue` public attribute, if the latter has previously been
defined.
'''
        return self.getpub('missing_value')
    #--- End: def
    @missing_value.setter
    def missing_value(self, value):
        self.setpub('missing_value', value)
        try:
            self.getpub('_FillValue')
            self.setpub('_FillValue', value)
        except AttributeError:
            pass
    #--- End: def
    @missing_value.deleter
    def missing_value(self):  
        self.delpub('missing_value')
    #--- End: def

    # --------------------------------------------------------
    # Read-only property attributes
    # --------------------------------------------------------
    @property
    def privattrs(self):
        '''
        
Return a dictionary of the instance's private attributes.
        '''
        try:
            return self._private['privattrs'].copy()
        except (KeyError, AttributeError):
            return {}
    #--- End: def

    @property
    def pubattrs(self):
        '''

Return a dictionary of the instance's public attributes.
'''
        out = {}

        # Special cases which don't live in the
        # self._private['pubattrs'] dictionary
        for attr in ('units', 'calendar', '_FillValue'):
            try:
                out[attr] = getattr(self, attr)
            except AttributeError:
                pass
        #--- End: for
        
        try:
            out.update(self._private['pubattrs'].copy())
        except (KeyError, AttributeError):
            pass

        return out
    #--- End: def

    @property
    def slice(self):
        '''

Slice the variable's data, returning a new variable.

**Examples:**

>>> v
<CF Variable: air_temperature(73, 96)>
>>> v.slice[0,0]
<CF Variable: air_temperature(1,1)>
'''
        return SliceVariable(self)
    #--- End: def

    @property
    def shape(self):
        '''

    Tuple of the data's dimension sizes.

    :returns: A tuple

    '''
        try:
            shape = self._data.shape
        except AttributeError:
            raise AttributeError, "Can't find 'shape' of '%s' instance" % \
                self.__class__.__name__

        return shape
    #--- End: def

    @property
    def ndim(self):
        '''

Number of data dimensions.

Equivalent to the number of elements in shape tuple.

:returns:     A non-negative integer
'''
        try:
            return self._data.ndim
        except AttributeError:
            raise AttributeError, "Can't find 'ndim' of '%s' instance" % \
                self.__class__.__name__
    #--- End: def

    @property
    def size(self):
        '''

    Number of elements in the data.

    Equivalent to the product of the data's dimension sizes.

    :returns:     A non-negative integer

    '''
        try:
            return self._data.size
        except AttributeError:  
            raise AttributeError, "Can't find 'size' of '%s' instance" % \
                self.__class__.__name__
    #--- End: def

    @property
    def dtype(self):
        '''

Data-type of the data's elements.

:returns:     A numpy dtype object.
'''
        try:
            return self._data.dtype
        except AttributeError:
            raise AttributeError, "Can't find 'dtype' of '%s' instance" % \
                self.__class__.__name__
    #--- End: def

    @property
    def array(self):
        '''

Create a numpy array deep copy of the data.

If the data was stored as a file pointer then it will be changed
in-place to be stored as a numpy array.

:returns:     A numpy array.

**Examples:**

>>> a = x.array
>>> type(a)
<type 'numpy.ndarray'>
>>> a = x.array[[0,1]]
>>> type(a)
<type 'numpy.ndarray'>
>>> a.shape
(2,)

'''
        if len(self._data) == 1:
            self._data.conform()
            return self._data._list[0].copy()
        else:
            data = self._data.concatenate()
            return data._list[0]
    #--- End: def

    @property
    def varray(self):
        '''
Create a numpy view of the data.

If the data was stored as a file pointer then it will be changed
in-place to be stored as a numpy array.

Note that making changes to elements of the returned view changes the
underlying data. Refer to :obj:`numpy.ndarray.view`.
    
:returns:     A numpy view.

**Examples:**

>>> a = x.varray
>>> print a
array([0, 1, 2, 3, 4])
>>> a[0] = 999
>>> print x.varray[0]
999
>>> a = 'a_different_object'
>>> print x.varray
array([999, 1, 2, 3, 4])
'''       
        self._concatenate_data()
        return self._data._list[0].view()
    #--- End: def

#    @property
#    def tarray(self):
#        if not self.istime():
#            raise TypeError, "Can't create a 'Time' array from a non-time variable"
#            
#        try:
#            calendar = self.calendar
#        except AttributeError:
#            calendar = 'standard'
#
#        return cf.Time(self.varray, units=self.units, calendar=calendar)
#    #--- End: def
#
#    def set_time(self, cftime):
#        if not self.istime():
#            raise TypeError, "Can't set time for a non-time variable"
#
#        # Check that the Time array has the same size and shape as the
#        # Variable array
#
#        self.units    = cftime.units
#        self.calendar = cftime.calendar
#
#        self[...] = cftime.array
#    #--- End: def

    def istime(self):
        '''
'''
        try:
            if ' since ' in self.units:
                return True
            return False
        except AttributeError:
            return False
    #--- End: def

#    def change_calendar(self, calendar, fix=False):
#        t = self.tarray
#        t = t.change_calendar(calendar, fix=fix)
#        self.set_time(t)
#    #--- End: def
        
    @property
    def first_datum(self):
        ''' 

    Return the first element of the data without replacing a file
    pointer with a numpy array.
        
    :returns:     The scalar value of the first element of the data.

    '''
        return numpy.asscalar(self.slice[(slice(0,1),) * self.ndim].varray)
    #--- End: def
    
    @property
    def last_datum(self):
        ''' 

    Return the last element of the data without replacing a file
    pointer with a numpy array.
    
    :returns:     The scalar value of the last element of the data.

    '''
        return numpy.asscalar(self.slice[(slice(-1,None),) * self.ndim].varray)
    #--- End: def

    def type(self):
        '''

The types of the data stored in the variable's Data object.

:returns: A list of type objects for each of the data elements.

 **Examples:**

>>> v.type()
[<type 'netCDF4.Variable'>]
>>> v.slice[...]
>>> v.type()
[<type 'numpy.ndarray'>]
'''
        try:
            return map(type, self._data._list)
        except AttributeError:
            raise AttributeError, "Can't find types of '%s' data" % \
                self.__class__.__name__
    #--- End: def

    def copy(self, data=True):
        '''

Return a deep copy.

:keyword data: *Optional*. If False then assign the output variable's
               data elements to None, with the expectation that they
               will be replaced subsequently. The data object's
               metadata are deep copied in any case.

:type data: bool

:returns: A deep copy.
'''
        new = type(self)()

        try:
            new._data = self._data.copy(data=data)
        except AttributeError:
            pass

        # private 
        attrs = set((self._private)) - set(('_data',))
        for attr in attrs:
            new._private[attr] = copy.deepcopy(self._private[attr])

        # Attributes
        attrs = set(self.__dict__) - set(('_private',))
        for attr in attrs:
            setattr(new, attr, copy.deepcopy(self.__dict__[attr]))

        return new
    #--- End: def

    def dump(self, id=None, omit=()):
        '''

Return a string containing a full description of the variable.

:keyword id: *Optional*. Set the common prefix of component names. If
             None then defaults to the class name.

:type id: str

:keyword omit: *Optional*. Omit the given attributes from the
               description.

:type omit: sequence

:returns: A string containing the description of the variable.

.. seealso:: :func:`cf.dump`
'''
        if id is None:
            id = self.__class__.__name__
            
        string = []
  
        try:
            string.append('%s.shape = %s' % (id, self.shape))
        except AttributeError:
            pass
        else:   
            if self.ndim <= 1:
                if self.size == 1:
                    string.append('%s.slice[...] = %s' % (id, self.first_datum))
                else:
                    string.append('%s.slice[[0,-1]] = %s' % \
                                      (id, numpy.array([self.first_datum,  
                                                        self.last_datum])))
            else:
                f = (0,)  * self.ndim
                l = (-1,) * self.ndim
                fs = ','.join([' 0'] * self.ndim)
                ls = ','.join(['-1'] * self.ndim)
                string.append('%s.slice[%s] = %s' % (id, fs, self.first_datum))
                string.append('%s.slice[%s] = %s' % (id, ls, self.last_datum))
        #--- End: try

        string.append('')

        for attr in sorted(set(self.pubattrs) - set(omit)):
            name  = '%(id)s.%(attr)s = ' % locals()
            value = repr(self.pubattrs[attr])
            indent = ' ' * (len(name))
            if value.startswith("'") or value.startswith('"'):
                indent = '%(indent)s ' % locals()
            string.append(textwrap.fill(name+value, 79,
                                        subsequent_indent=indent))
        #--- End: for

        string.append('')

        return '\n'.join(string)
    #--- End: def

    def equals(self, other, rtol=None, atol=None):
        '''

Return True if the variable is congruent to another variable in that

    1. They have identical data. Pointers to data on disk may be
       converted to numpy arrays in memory if required.

    2. They have identical reserved public attributes.

    3. They have identical public attributes.

Equality of numbers is to within a tolerance. Refer to :mod:`cf` for
details.

:argument other: The variable to compare against for equality.

:type other: object
             
:keyword atol: *Optional*. If None then use the default method for
               setting the absolute tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the absolute tolerance to this value for all such
               comparisons.

:type atol: None or float

:keyword rtol: *Optional*. If None then use the default method for
               setting the relative tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the relative tolerance to this value for all such
               comparisons.

:type rtol: None or float

:returns: True if the two objects are congruent, False otherwise.


**Examples:**

>>> y = x
>>> x.equals(y)
True
>>> y = x + 1
>>> x.equals(y)
False
>>> y -= 1
>>> x.equals(y)
True
>>> x.setpub('name', 'name0')
>>> y.setpub('name', 'name1')
>>> x.equals(y)
False
'''
        # Check each instance's id
        if id(self) == id(other):
            return True

        # Check that each instance is the same type
        if type(self) != type(other):
            return False

        self_pub  = self.pubattrs
        other_pub = other.pubattrs
        if set(self_pub) != set(other_pub):
            return False
       
        self_private  = self._private
        other_private = other._private
        if set(self_private) != set(other_private):
            return False
       
        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()                
            
        # Check that each instance has the same CF property values
        for attr, x in self_pub.iteritems():
            y = other_pub[attr]
            if not equals(x, y, rtol=rtol, atol=atol):
                return False

        # Check the private attributes which have equals methods, such
        # as _data, units, etc.
        for attr, x in self_private.iteritems():
            y = other_private[attr]
            try:
                if not x.equals(y, rtol=rtol, atol=atol):
                    return False
            except AttributeError:
                pass
        #--- End: for

        return True
    #--- End: def

    def _match(self, exact, **kwargs):
        '''

See match and ematch for details.
'''        
        # Loop round criteria
        for name, value in kwargs.iteritems():

            if value is None:
                continue

            # ------------------------------------------------------
            # Find out how many phenomena match 'name'
            # ------------------------------------------------------
            count = 0
             
            # Check scalar or 1-d, size 1 coordinates, if this
            # variable has them
            try:
                x = self.coord(value, exact=exact, key=True)
            except AttributeError:
                pass
            else:
                try:
                    if x.size == 1 and x.ndim <= 1:
                        x = x.first_datum                
                        count += 1
                except AttributeError:
                    pass

            # Check public attributes
            attr = None
            if exact:
                if name in self.pubattrs:
                    attr = name
                    count += 1
            else:                    
                for public in self.pubattrs:
                    if re.match(name, public):
                        attr = public
                        count += 1
                        
            if count == 1:
                # Variable has exactly one phenomenon name matching 'name'
                if attr is not None:
                    x = self.getpub(attr)
            elif count == 0:
                # Variable has no phenomenon name matching 'name'
                return False
            elif count > 1:                
                raise ValueError, "More than one phenomenon name matches '%(value)s'" % \
                    locals()
    
            # --------------------------------------------------------
            # See if the phenomenon conforms with the value
            # --------------------------------------------------------
            try:
                if not value.evaluate(x):
                    return False
                continue
            except AttributeError:
                pass

            if not isinstance(value, (list, tuple, numpy.ndarray)):
                values = (value,)
            else:
                values = value

            for v in values:  
                try:
                    if re.search('^%s$' % v, x):
                        break
                except TypeError:
                    # v is not a string
                    if equals(x, v):
                        break

                # Still here?
                return False
            #--- End: for
        #--- End: for

        # Still here? Then we have a match
        return True
    #--- End: def
      
    def match(self, **kwargs):      
        '''

Determine whether or not a variable matches conditions on its
phenomena, i.e. its public attributes and, if it has any, its scalar
or 1-d, size 1 coordinates. Unambiguous abbreviations of public
attribute or coordinate names are allowed.

The phenomena and their conditions are specified with `**kwargs`
parameters.

The variable matches the conditions if and only if it contains all of
the specified phenomena and they satisfy all of their given
criteria. A variable always matches no criteria.

:keyword **kwargs: The keyword arguments identify a public attribute's
                   name or a coordinate's standard name. Any
                   unambiguous abbreviation of the phenomenon's name
                   is acceptable.
          
                   A keyword's value may be one of the following:
                   
                   1. A number. A match requires that a numeric valued
                      phenomenon equals the number using the
                      numerically tolerant equality.
                   
                   2. A string. A match requires that a string-valued
                      phenomenon passes a regular expression match on
                      the string. The string may contain regular
                      expression special characters. To avoid
                      ambiguities, it is assumed that the regular
                      expression string matches the whole of the
                      phenomenon string, i.e. the regular expression
                      special characters ^ and $ are assumed if not
                      given.
                   
                   3. A sequence of numbers or strings. A match
                      requires that a numeric phenomenon equals,
                      within numerical tolerance, at least one of the
                      sequence's elements or a string-valued
                      phenomenon passes a regular expression match for
                      at least one string in the sequence (see 2.).
                   
                   4. A :class:`~cf.Comparison` object. A match
                      requires that the comparison for the phenomenon
                      evaluates to True.

:returns: True if the variable matches the criteria, False otherwise.
        
**Examples:**

>>> s
<CF Space: air_temperature(19, 30, 24)>
>>> s.pub('standard_name'), s.pub('units')
('air_temperature', 'K')
>>> s.extract(standard_name = 'air_temperature')
True
>>> s.extract(standard = '^air.*')
True
>>> s.extract(standard_name = lt('b'))
True
>>> s.extract(standard_name = outside('b', 'p'))
True
>>> s.extract(standard = ['.*temp.*', 'pressure'])
True
>>> s.extract(height = 2.0, units = 'K')
True
'''
        return self._match(False, **kwargs)
    #--- End: def

    def ematch(self, **kwargs):
        '''

Determine whether or not a variable matches conditions on its
phenomena, i.e. its public attributes and, if it has any, its scalar
or 1-d, size 1 coordinates. As :meth:`~cf.Variable.match`, but with
the restriction that only exact public attribute or coordinate names
are allowed.

:keyword **kwargs: The keyword arguments identify exact phenomenon
                   names, either a public attribute names or 1-d, size
                   1 coorinate standard names.

                   The keyword values are as for
                   :meth:`~cf.Variable.match`.

:returns: True if the variable matches the criteria, False otherwise.
        
**Examples:**

>>> s
<CF Space: air_temperature(19, 30, 24)>
>>> s.standard_name
'air_temperature'
>>> s.extract(standard_name = 'air_temperature')
True
>>> s.extract(standard_n = '^air.*')
False
'''
        return self._match(True, **kwargs)
    #--- End: def

    def extract(self, **kwargs):
        '''

Return the variable if and only if it matches conditions on its
phenomena, i.e. its public attributes and, if it has any, its scalar
or 1-d, size 1 coordinates. Unambiguous abbreviations of public
attribute or coordinate names are allowed.

Refer to :meth:`~cf.Variable.match` for further details.

keyword **kwargs: The keyword arguments identify a public attribute's
                  name or a coordinate's standard name. Any
                  unambiguous abbreviation of the phenomenon's name is
                  acceptable.

                  The keyword values are as for
                  :meth:`~cf.Variable.match`.

:returns: The variable if the it matches the criteria, otherwise a
          ValueError is raised.
'''
        if self.match(**kwargs):
            return self
        else:
            raise ValueError, "dddhkjhjsdkj"
    #--- End: def

    def eextract(self, **kwargs):
        '''

Return the variable if and only if it matches conditions on its
phenomena, i.e. its public attributes and, if it has any, its scalar
or 1-d, size 1 coordinates. As :meth:`~cf.Variable.extract`, but with
the restriction that only exact public attribute or coordinate names
are allowed.

keyword **kwargs: The keyword arguments identify exact phenomenon
                  names, either a public attribute names or 1-d, size
                  1 coorinate standard names.

                  The keyword values are as for
                  :meth:`~cf.Variable.match`.

:returns: The variable if the it matches the criteria, otherwise a
          ValueError is raised.
'''
        if self.ematch(**kwargs):
            return self
        else:
            raise ValueError, "5555dddhkjhjsdkj"
    #--- End: def

    def priv(self, *args, **kwargs):
        '''
        
A method which may be used in place of :meth:`~cf.Variable.getpriv` or
:meth:`~cf.Variable.setpriv`.

If called with positional parameters, then they are interpreted as for
:meth:`~cf.Variable.getpriv` and a public attribute (or default) value
is returned.

If called with a keyword parameter, then the keyword and its value act
as the two positional parameters of :meth:`~cf.Variable.setpriv` and a
public attribute is set.
'''
        if args:
            if kwargs:
                raise TypeError, "sdfsdfsd 0u812 3"            
            return self.getpriv(*args)

        for attr, value in kwargs.iteritems():
            self.setpriv(attr, value)
    #--- End: def

    def setpriv(self, attr, value):
        '''
Set a private attribute.

:keyword attr: The name of the private attribute to be set.

:type attr: str

:keyword value: *Optional*. The value for the private attribute.
                 
:type value: object

:returns: None

'''
        try:
            self._private['privattrs'][attr] = value
        except KeyError:
            self._private['privattrs'] = {attr: value}       
    #--- End: def

    def getpriv(self, attr, *default):
        '''
Return a private attribute.

:keyword attr: The name of the private attribute to be returned.

:type attr: str

:keyword default: *Optional*. Return `default` if the requested private
                   attribute has not been set
                 
:type default: object

:returns: The value of a private attribute.
'''
        try:
            return self._private['privattrs'][attr]
        except KeyError:
            if default:
                return default[0]

        raise AttributeError, "'%s' object doesn't have private attribute '%s'" % \
            (self.__class__.__name__, attr)          
    #--- End: def

    def delpriv(self, attr):
        '''
Delete a private attribute.

:keyword attr: The name of the private attribute to be deleted.

:type attr: str

:returns: None
'''
        try:
            del self._private['privattrs'][attr]
        except KeyError:
            raise AttributeError, "'%s' object has no private attribute '%s'" % \
                (self.__class__.__name__, attr)
    #--- End: def

    def pub(self, *args, **kwargs):
        '''
        
A method which may be used in place of :meth:`~cf.Variable.getpub` or
:meth:`~cf.Variable.setpub`.

If called with positional parameters, then they are interpreted as for
:meth:`~cf.Variable.getpub` and a public attribute (or default) value
is returned.

If called with a keyword parameter, then the keyword and its value act
as the two positional parameters of :meth:`~cf.Variable.setpub` and a
public attribute is set.


**Examples:**

>>> v.pub(units='K')
>>> v.pub('units')
'K'
>>> v.pub('UNITS', 'default_value')
'default_value'
'''
        if args:
            if kwargs:
                raise TypeError, "sdfsdfsd 0u812"            
            return self.getpub(*args)

        for attr, value in kwargs.iteritems():
            self.setpub(attr, value)
    #--- End: def

    def setpub(self, attr, value):
        '''
Set a public attribute.

:keyword attr: The name of the public attribute to be set.

:type attr: str

:keyword value: *Optional*. The value for the public attribute.
                 
:type value: object

:returns: None
'''
        # Special cases which don't live in the
        # self._private['pubattrs'] dictionary            
        if attr in ('units', 'calendar', '_FillValue'):
            setattr(self, attr, value)
            return None

        # Still here?
        try:
            self._private['pubattrs'][attr] = value
        except KeyError:
            self._private['pubattrs'] = {attr: value}
    #--- End: def

    def getpub(self, attr, *default):
        '''

Return a public attribute.

:keyword attr: The name of the public attribute to be returned.

:type attr: str

:keyword default: *Optional*. Return `default` if the requested public
                   attribute has not been set
                 
:type default: object

:returns: The value of a public attribute.
'''
        # Special cases which don't live in the
        # self._private['pubattrs'] dictionary
        if attr in ('units', 'calendar', '_FillValue'):
            return getattr(self, attr, *default)
        
        # Still here?
        try:
            return self._private['pubattrs'][attr]
        except (KeyError, AttributeError):
            if default:
                return default[0]

        # Still here?
        raise AttributeError, "'%s' object doesn't have public attribute '%s'" % \
            (self.__class__.__name__, attr)          
    #--- End: def

    def delpub(self, attr):
        '''
Delete a public attribute.

:keyword attr: The name of the public attribute to be deleted.

:type attr: str

:returns: None
'''
        # Special cases which don't live in the
        # self._private['pubattrs'] dictionary
        if attr in ('units', 'calendar', '_FillValue'):
            delattr(self, attr)
            return None

        # Still here?
        try:
            del self._private['pubattrs'][attr]
        except (KeyError, AttributeError):
            raise AttributeError, "'%s' object doesn't have public attribute '%s'" % \
                (self.__class__.__name__, attr)
    #--- End: def

    def name(self, long=False, ncvar=False, default=None):
        ''' 

Return the `standard_name` attribute of the variable.

If there is no `standard_name` attribute then return one of the
`long_name` attribute, the `ncvar` attribute or the value of the
default parameter, depending on the values of the parameters.

:keyword long: *Optional*. If True, return the `long_name` if
               `standard_name` does not exist.

:type long: bool

:keyword ncvar: *Optional* If True, return `ncvar` if neither
                `standard_name` not `long_name` have already been
                returned.

:type ncvar: bool

:keyword default: *Optional*. Return `default` if `standard_name`,
                  `long_name` nor `ncvar` have already been returned.

:type default: None or object


:returns: The name of the variable.

'''
        try:
            return self.getpub('standard_name')
        except AttributeError:
            pass

        if long:
            try:
                return self.getpub('long_name')
            except AttributeError:
                pass
            
        if ncvar:
            try: 
                return self.getpriv('ncvar')
            except AttributeError: 
                pass            

        return default
    #--- End: def
        
#--- End: class


class Coordinate(Variable):
    '''

A CF coordinate object.

Refer to the :class:`~cf.Variable` and :class:`~cf.Grid` classes and
the :mod:`cf` package for details.


**Reserved public attributes:**

In addition to the reserved public attributes inherited from the
:class:`~cf.Variable` class, the following are also reserved public
attributes:

===================== ========================
Attribute             Description             
===================== ========================
:attr:`axis`          As in the CF convention.
:attr:`positive`      As in the CF convention.
===================== ========================


**Reserved attributes:**

In addition to the reserved attributes inherited from the
:class:`~cf.Variable` class, the following attributes have special
definitions and it is recommended to set them accordingly.

=================== =================================================
Attribute           Description
=================== =================================================
:attr:`bounds`      The coordinate's :class:`~cf.CoordinateBounds`
                    object.

:attr:`climatology` An indicator to specify if the coordinate's
                    bounds are intervals of climatological time.

:attr:`transform`   A pointer to the coordinate's
                    :class:`~cf.Transform` object.
=================== =================================================

'''

    @property
    def slice(self):
        '''

Slice the coordinate's data, returning a new coordinate. The
coordinate' bounds are also sliced.

**Examples:**

>>> c
<CF Coordinate: longitude(96)>
>>> c.bounds
<CF CoordinateBounds: longitude(96,2)>
>>> d=c.slice[1:]
>>> d
<CF Coordinate: longitude(95)>
>>> d.bounds
<CF CoordinateBounds: longitude(95,2)>
'''
        return SliceCoordinate(self)
    #--- End: def
   
    # ----------------------------------------------------------------
    # Property attribute: axis
    # ----------------------------------------------------------------
    @property
    def axis(self):
        '''

The coordinate's CF `axis` public attribute. 
'''
        return self.getpub('axis')
    #--- End: def
    @axis.setter
    def axis(self, value): self.setpub('axis', value)
    @axis.deleter
    def axis(self):        self.delpub('axis')

    # ----------------------------------------------------------------
    # Property attribute: positive
    # ----------------------------------------------------------------
    @property
    def positive(self):
        '''

The coordinate's CF `positive` public attribute. 
'''
        return self.getpub('positive')
    #--- End: def
    @positive.setter
    def positive(self, value): self.setpub('positive', value)
    @positive.deleter
    def positive(self):        self.delpub('positive')

    # ----------------------------------------------------------------
    # Property attribute: bounds (not a public attribute)
    # ----------------------------------------------------------------
    @property
    def bounds(self):
        '''

The coordinate's bounds object.
        '''
        return self._getter('bounds')
    #--- End: def
    @bounds.setter
    def bounds(self, value): self._setter('bounds', value)
    @bounds.deleter
    def bounds(self):        self._deleter('bounds')

    # ----------------------------------------------------------------
    # Property attribute: climatology (not a public attribute)
    # ----------------------------------------------------------------
    @property
    def climatology(self):
        '''

Indicator of whether the coordinate's bounds represent climatologocal
time intervals.
        '''
        return self._getter('climatology')
    #--- End: def
    @climatology.setter
    def climatology(self, value): self._setter('climatology', value)
    @climatology.deleter
    def climatology(self):        self._deleter('climatology')
   
    # ----------------------------------------------------------------
    # Property attribute: transform (not a reserved public attribute)
    # ----------------------------------------------------------------
    @property
    def transform(self):
        '''

Pointer to coordinate's transfrom.
        '''
        return self._getter('transform')
    #--- End: for
    @transform.setter
    def transform(self, value): self._setter('transform', value)
    @transform.deleter
    def transform(self):        self._deleter('transform')

    def dump(self, id=None, omit=()): 
        '''

Return a string containing a full description of the coordinate.

:keyword id: *Optional*. Set the common prefix of component names. If
             None then defaults to the class name.

:type id: str

:keyword omit: *Optional*. Omit the given attributes from the
               description.

:type omit: sequence


:returns: A string containing the description of the coordinate.

.. seealso:: :func:`cf.dump`
'''
        if id is None:
            id = self.__class__.__name__       

        default = self.getpriv('ncvar', '')

        string = ['%s coordinate' % self.name(default=default, long=True)]
        string.append(''.ljust(len(string[0]), '-'))
        
        string.append(super(Coordinate,self).dump(id=id, omit=omit))

        try:
            bounds = self.bounds
        except AttributeError:
            pass
        else:
            try:
                bounds._data
            except AttributeError:
                string.append('%s.bounds = %s' % (id, repr(bounds)))
            else:
                string.append('%s.bounds.shape = %s' % (id, bounds.shape))
                if bounds.ndim == 1:
                    string.append('%s.bounds[...] = %s' % \
                                      (id,
                                       numpy.array((bounds.first_datum,
                                                    bounds.last_datum))))
                else:
                    ndim_minus1 = bounds.ndim - 1
                    f = (0,)  * ndim_minus1
                    l = (-1,) * ndim_minus1
                    f += (slice(None),)
                    l += (slice(None),)
                    
                    fs = ','.join([' 0'] * ndim_minus1)
                    ls = ','.join(['-1'] * ndim_minus1)

                    string.append('%s.bounds.slice[%s,:] = %s' % (id, fs, bounds._data.slice[f].varray))
                    string.append('%s.bounds.slice[%s,:] = %s' % (id, ls, bounds._data.slice[l].varray))
            #--- End: try
            try:
                string.append('%s.climatology = %s' % (id, self.climatology))
            except AttributeError:
                pass
            string.append('')
        #--- End: try

        try:
            string.append('%s.transform = %s\n' % (id,
                                                   repr(self.transform)))
        except AttributeError:
            pass

        return '\n'.join(string)
    #--- End: def

    def _data_insert(self, indices, other,
                     self_dimensions, other_dimensions, 
                     match_for_dim1):
        '''
'''
        # ------------------------------------------------------------
        # Check that either both coordinates have bounds or neither
        # coordinate has bounds
        # ------------------------------------------------------------
        try:
            self_bounds = self.bounds
        except AttributeError:
            try:
                other.bounds
                raise AttributeError, "Can't insert coordinate data: Missing bounds"
            except AttributeError:
                # Neither coordinate has bounds
                pass
        else:
            try:
                other_bounds = other.bounds
                # Both coordinates have bounds
            except AttributeError:
                raise AttributeError, "Can't insert coordinate data: Missing bounds"
        #--- End: try

        # Insert the coordinate's data
        super(Coordinate, self)._data_insert(indices, other,
                                             self_dimensions,
                                             other_dimensions,
                                             match_for_dim1)

        try:
            other_bounds
        except NameError:
            pass
        else:
            # Insert the coordinate's bounds
            self.bounds._data_insert(indices, other.bounds,
                                     self_dimensions+['bounds'], 
                                     other_dimensions+['bounds'], 
                                     match_for_dim1)        
        #--- End: try

    #--- End: def

    def _concatenate_data(self):
        '''

Concatenate a coordinate's data
'''
#        print 'dch _concatenate_data ',self.__class__.__name__
        super(Coordinate, self)._concatenate_data()
        try: 
            # Concatenate the coordinate's bounds
            self.bounds._concatenate_data()
        except AttributeError:
            # The coordinate knows no bounds
            pass
    #--- End: for
 
#--- End: class

class CoordinateBounds(Variable):
    '''
    
A CF coordinate's bounds object containing cell boundaries or
intervals of climatological time. The containing coordinate's
`climatology` attribute indicates which type of bounds are present.

Refer to the parent :class:`~cf.Variable` class for all details.

'''
    pass
#--- End: class


class CellMeasures(Variable):
    '''
    
A CF cell measures object containing information that is needed about
the size, shape or location of the spaces's cells.

Refer to the parent :class:`~cf.Variable` class for all details.
'''
    pass
#--- End: class


class Space(Variable):
    '''

Represent a space construct according to the CF data model.

Refer to the :class:`~cf.Variable` class for details.

Refer to the :mod:`cf` module for details on data access by indexing
and data storage.


:keyword public_attributes: *Optional*. The new instance is given the
                            public attributes named and valued from
                            the key, value pairs of this
                            dictionary. The values are deep copied.

:type public_attributes: dict

:keyword private_attributes: *Optional*. The new instance is given the
                             private attributes named and valued from
                             the key, value pairs of this
                             dictionary. The values are deep copied.

:type private_attributes: dict


**Reserved attributes:**

In addition to the reserved attributes inherited from the
:class:`~cf.Variable` class, the following attributes have special
definitions and it is recommended to set them accordingly.

==================== ============================================
Attribute            Description             
==================== ============================================
:attr:`cell_methods` The space's :class:`~cf.CellMethods` object.
:attr:`grid`         The space's :class:`~cf.Grid` object.
==================== ============================================

**Methods and attributes defined here:**
'''

#    _reserved_public_attrs = Variable._reserved_public_attrs
#    _reserved_public_attrs += ('cell_methods',)
    
    def __str__(self):
        '''
x.__str__() <==> str(x)
'''
        string = []

        try:
            grid = self.grid
        except AttributeError:
            # Space has no grid
            dim_names = self.shape
        else:     
            # Space has a grid
            dim_names = []
            for dim in grid.dimensions['data']:
                try:
                    dim_names.append(grid[dim].name(ncvar=True, 
                                                    default=dim))
                except (KeyError, AttributeError):
                    try:                        
                        dim_names.append(grid.nc_dimensions[dim])
                    except (KeyError, AttributeError):
                        dim_names.append(dim)
            #--- End: for
            dim_names = tuple(dim_names)
        # ---End: try

        dim_names = str(dim_names).replace("'", "")
        dim_names = dim_names.replace(',)', ')', 1)

        string = ['Data            : %s%s' %(self.name(long=True, 
                                                       ncvar=True, 
                                                       default=''),
                                             dim_names)]
        
        # Cell methods
        try:
            string.append(str(self.cell_methods))
        except AttributeError:
            # Space has no cell methods
            pass

        # Grid
        try:
            string.append(str(grid))
        except NameError:
            # Space has no grid
            string.append('Dimensions      : \nAuxiliary coords: \nCell measures   : ')

        return '\n'.join(string)
    #--- End def

    def _concatenate_data(self):
        '''

Concatenate a space's data
'''
        original_n_elements = len(self._data)

        # Concatenate the data elements of the field
        super(Space, self)._concatenate_data()

        # Concatenate the data elements of the grid components which
        # share a dimension with the field's data
        try:
            grid = self.grid
        except AttributeError:
            pass
        else:
            if original_n_elements > 1:
                for key in grid:
                    grid[key]._concatenate_data()
        #--- End: try

    #--- End: def

    # ----------------------------------------------------------------
    # Property attribute: ancillary_variables
    # ----------------------------------------------------------------
    @property
    def ancillary_variables(self):        return self._getter('ancillary_variables')
    @ancillary_variables.setter
    def ancillary_variables(self, value): self._setter('ancillary_variables', value)
    @ancillary_variables.deleter
    def ancillary_variables(self):        self._deleter('ancillary_variables')

    # ----------------------------------------------------------------
    # Property attribute: cell_methods
    # ----------------------------------------------------------------
    @property
    def cell_methods(self): 
        '''

The space's CellMethods object.
        '''
        return self.getpub('cell_methods')
    @cell_methods.setter
    def cell_methods(self, value): self.setpub('cell_methods', value)
    @cell_methods.deleter
    def cell_methods(self):        self.delpub('cell_methods')

    # ----------------------------------------------------------------
    # Property attribute: grid
    # ----------------------------------------------------------------
    @property
    def grid(self):
        '''

The space's Grid object
        '''
        return self._getter('grid') 
    #--- End: def
    @grid.setter
    def grid(self, value): self._setter('grid', value)
    @grid.deleter
    def grid(self):        self._deleter('grid')

    def reverse_axes(self, *names):
        '''

Reverse the direction of named axes *IN PLACE*. The reverse does not
actually occur until the data are accessed.
'''
        dimensions = self.grid.dimensions.copy()
        data_dims  = dimensions.pop('data')

        for name in names:
            dim = self.coord(name, key=True)

            try:
                i = data_dims.index(dim)
            except ValueError:
                # Scalar
                pass
            else:
                try:
                    ddim = self._data.order[i]
                except AttributeError:
                    # No data
                    pass
                else:
                    # 1d or Nd
                    self._data.direction[ddim] = not self._data.direction[ddim]
                    try:
                        if self._data.adim is i:
                            self._data._list.reverse()
                            self._data._elements.reverse()
                    except AttributeError:
                        pass
            #--- End: try
        
            for key, dims in dimensions.iteritems():
                try:
                    i = dims.index(dim)
                except ValueError:
                   pass
                else:                
                    try:
                        ddim = self.grid[key]._data.order[i]
                    except IndexError:
                        # Scalar
                        self.grid[key]._data.direction = \
                            not self.grid[key]._data.direction
                    except AttributeError:
                        # No data
                        pass
                    else:
                        # 1d or Nd
                        self.grid[key]._data.direction[ddim] = \
                            not self.grid[key]._data.direction[ddim]
                        try:
                            if self.grid[key]._data.adim is i:
                                self.grid[key]._data._list.reverse()
                                self.grid[key]._data._elements.reverse()
                        except AttributeError:
                            pass
                #--- End: try
            #--- End: for
        #--- End: for
    #--- End: def

    def squash(self):
        '''

Squash axes *IN PLACE*. I.e. squeeze size 1 dimensions from the space
and its dimension coordinates if and only if a size 1 dimension is not
spanned by an auxuliary coordinate or a cell measures variable.
'''

        for dim in self.grid:
            # --------------------------------------------------------
            # Only consider squeezing this dimension if it is a 1d,
            # size 1 dimension coordinate
            # --------------------------------------------------------
            if not dim.startswith('d'):
                continue

            c = self.grid[dim]._data
            if not (c.size == 1 and c.ndim == 1):
                continue

            # --------------------------------------------------------
            # Do not squeeze this dimension if it is being used by an
            # auxiliary coordinate or cell measures
            # --------------------------------------------------------
            squeeze = True
            for key in self.grid:
                if key.startswith('d'):
                    continue
                if dim in self.grid.dimensions[key]:
                    squeeze = False
                    break
            #--- End: for
            if not squeeze:
                continue

            # --------------------------------------------------------
            # Convert the dimension coordinate to a scalar
            # --------------------------------------------------------
            c.shape                 = ()            
            c.order                 = []
#            c.direction             = {}        
            c.direction             = c.direction.values()[0] # c.direction  -> scalar bool

            c.elements[0].order     = []
            c.elements[0].direction = []
            c.elements[0].shape     = []
            try:
                b = self.grid[dim].bounds._data
            except AttributeError:
                pass
            else:
                b.shape                 = (2,)                
                b.order                 = ['bounds']
                b.direction             = {'bounds': b.direction['bounds']}
                b.elements[0].order     = ['bounds']
                b.elements[0].shape     = [2]
                b.elements[0].direction.pop(0)
            #--- End: try

            # --------------------------------------------------------
            # Drop this dimension from the field's data
            # --------------------------------------------------------
            try:
                index = self.grid.dimensions['data'].index(dim)
            except ValueError:
                pass
            else:
                self.grid.dimensions['data'].pop(index)
                self._data.remove_dimension(index)
        #--- End: for
    #--- End: def

    def dump(self, id=None, omit=()):
        '''

Return a string containing a full description of the space.

:keyword id: *Optional*. Set the common prefix of variable
             component names. If None then defaults to the class
             name.

:keyword id: str

:keyword omit: *Optional*. Omit the given attributes the
               description.

:type omit: sequence


:returns: A string containing the description of the space.

.. seealso:: :func:`cf.dump`
'''
        if id is None:
            id = self.__class__.__name__

        # Produce summary
        string = ['space summary\n-------------']
        string.append(str(self))
        string.append('')

        # Produce dump
        name = self.name(default='', long=True)
        string.append(name)
        string.append(''.ljust(len(name), '-'))
        
        string.append(super(Space, self).dump(id=id, omit=('grid',)))
#        string.append(super(Space, self).dump(self, id=id, omit=omit))

        try:
            string.append(self.grid.dump(id='%(id)s.grid' % locals()))
        except AttributeError:
            pass

        return '\n'.join(string)
    #--- End: def

    def coord(self, name, role=None, key=False, exact=False):
        '''

Find a space's coordinate by name.

The given `name` argument is an abbreviation for (or equal to if the
`exact` parameter is True) the standard name of desired coordinate If
the `key` parameter is True then return the coordinate's grid key name
instead of the Coordinate variable. If a coordinate does not have
`standard_name` attribute, then `name` will be matched against its
`ncvar` attribute is used.

Note that the returned coordinate is an object identity to the
coordinate stored in the grid so, for example, a coordinate's
attributes may be changed in-place as follows:

>>> s.coord('height').long_name
AttributeError: 'Coordinate' object has no attribute 'long_name'
>>> s.coord('hei').long_name = 'HEIGHT'
>>> s.coord('heigh').long_name
'HEIGHT'

Or a deep copy may be made with the coordinate's
:meth:`~cf.Coordinate.copy` method:

>>> h = s.coord('height').copy()

:argument name: The string to identify a coordinate by name.

:type name: str

:keyword exact: *Optional*. If True then assume that the value of the
                `name` argument is equal to exactly one coordinate's
                name.

:type exact: bool

:keyword key: *Optional*. Return the grid key name instead of the
              coordinate.

:type key: str

:keyword role: *Optional*. Restrict the search to coordinates of the
               given role. Valid values are 'dim' and 'aux', for
               dimension and auxiliary coordinate types
               respectively. If None then both types of coordinates
               will be searched.

:type role: str or None

  
:returns: If a coordinate has been identified, return either a
          :class:`~cf.Coordinate` instance or, if the `keys` parameter
          is True, a grid key name string. otherwise, return None.


**Examples:**

>>> s.coord('lon')
<CF Coordinate: longitude(128)>
>>> s.coord('lon', key=True)
'dim2'
>>> s.coord('lonX', key=True)
None
>>> s.coord('lon', exact=True)
None
>>> s.coord('longitude', exact=True)
<CF Coordinate: longitude(128)>
'''
        try:
            return self.grid.coord(name, role=role, key=key, exact=exact)
        except AttributeError:
            return None
    #--- End: def

    @property
    def slice(self):
        '''

Slice the space's data, returning a new space. The space's grid is
also sliced.

Slicing by index
----------------

Refer to the 'Copying and resizing' in the :mod:`cf` package.

Slicing by coordinate criteria
------------------------------

Coordinates are identified by keyword parameters of their standard
names. If the optional first positional argument `exact` evaluates to
`False` (the default) then unambiguous abbreviations of the standard
names are accepted, otherwise exact standard names are required:

>>> s.slice(longitude=0.0)
<CF Space: air_temperature(12, 64, 1)
>>> s.slice(lon=0.0)
<CF Space: air_temperature(12, 64, 1)
>>> s.slice(l=0.0)
ValueError: Can't identify coordinate from keyword 'l'
>>> s.slice(False, lon=0.0)
<CF Space: air_temperature(12, 64, 1)
>>> s.slice(True, lon=0.0)
ValueError: Can't identify coordinate from keyword 'lon'

The coordinate criteria are given by the keyword parameters
values. Four different methods are available:

1. A scalar. The dimension is sliced where coordinate equals (using
   the coordinate's == operator) the scalar value:

   >>> s.slice(lat=0.0)

2. A sequence. The dimension is sliced where the coordinate equals
   (using the coordinate's == operator) the sequence's elements. All
   elements of the sequence must be found in the coordinate, but the
   sequence may contain fewer elements than the coordinate's data:

   >>> s.slice(lon=[0.0, 2.8125])

3. A :class:`~cf.Comparison` object (as, for example, returned by the
   :func:`~cf.lt` function). The dimension is sliced where the
   comparison for the coordinate evaluates to True:

   >>> s.slice(lat=cf.lt(0.0))

4. An numpy array of booleans with the same shape as the
   coordinate. The dimension is sliced where elements of this array
   are True:

   >>> s.slice(lat=s.coord('lat')<0.0)
  
If a coordinate is not identified with a keyword parameter then its
dimension is returned unchanged.

The order in which coordinates are given in the parameter list is
irrelevant.

A call with no keyword parameters is equivalent to indexing with an
Ellipsis (...).


**Examples:**

>>> s.slice(lat=0.0, lon=0.0)
<CF Space: air_temperature(12, 1, 1)
>>> s.slice(lon=cf.lt(0.0), lon=[0, 2.8125])
<CF Space: air_temperature(12, 32, 2)
>>> s.slice(lon=[0, 1])
ValeuError: 'longitude' coordinate doesn't contain the value 1
>>> s.slice(lon=1)
ValueError: No indices found for 'longitude' coordinate
'''
        return SliceSpace(self)
    #--- End: def

#    def uncompress(self):
#        compressed_dims = []
#
#        for dim in self.grid.dimensions['data']:
#            try:
#                self.grid[dim].compress
#                compressed_dims.append(dim)
#            except (KeyError, AttributeError):
#                pass
#
#        new = copy.deepcopy(self)
#
#        # Return if the data are not compresed by gathering
#        if not compressed_dims:
#            return new        
#        
#        # Still here?
#        dimN = len(new.grid.dimension_sizes)
#        comN = len(new.grid.get('com'))
#
#        for com_dim in compressed_dims:
#
#            compressed_list = new.grid[com_dim].compress
#
#            com = 'com'+str(comN)
#            new.grid[com] = new.grid.pop(com_dim)
#            new.grid[com].compress = []
#
#            new.grid.dimensions['data'].remove(com_dim)
#            
#            new_dim_sizes = ()
#      
#            for unc in compressed_list:
#                dim = 'dim'+str(dimN)
#                new.grid[com].compress.append(dim)
#                new.grid[dim] = new.grid.pop(unc)
#
#                new.grid.dimensions.pop(unc)
#                new.grid.dimensions[dim]      = dim
#                new.grid.dimension_sizes[dim] = new.grid.dimension_sizes.pop(unc)
#               
#                new_dim_sizes += (new.grid.dimension_sizes[dim],)
#
#                dimN  += 1                
#
#            
#            if new.ndim == 1:
#                new._data = numpy.expand_dims(new.varray, 1)
#                size  = 1
#                shape = ()
#            else:
#                new._data = numpy.swapaxes(new.varray, pos, 0)
#                size  = numpy.prod(shape)
#                shape = new.shape[1:]
#                new._data = numpy.reshape(new._data, (new._data.shape[0], size))
#
#            uncompressed = numpy.ma.masked_all((numpy.prod(new_dim_sizes), size),
#                                               dtype=new.dtype)
#            try:
#                uncompressed.fill_value = new.missing_value
#                uncompressed.fill_value = new.missing_value
#            except AttribueError:
#                pass
##            uncompressed = numpy.empty(numpy.prod(new_dim_sizes), size)
##            uncompressed = uncompressed.fill(new.missing_value)
#
#            for i in xrange(new._data.shape[0]):
#                uncompressed[new.grid[com][i],:] = new._data[i,:]
#
#            uncompressed = numpy.swapaxes(uncompressed, 0, 1)
#            uncompressed = numpy.reshape(uncompressed, shape+new_dim_sizes)
#            
#            new._data = uncompressed
#
#            comN += 1
#        #--- End: for dim in compressed_dims:
#
#        return new
#    #--- End: defSpace.uncompress

#--- End: class


class Grid(CfDict):
    '''

A CF grid object defining a space's dimensionality, coordinates, cell
measures and transformations in a dictionary-like object. Refer to the
`Grid Structure` section of the :mod:`cf` module for details.

:argument *args: Keys and values are initialized exactly as for a
                 built-in dict.

:argument **kwargs: Keys and values are initialized exactly as for a
                    built-in dict.

**Overloaded operators**

The ``in`` (set membership) operator is overloaded to use numerically
tolerant equality.

**Special attributes:**

======================= ================================================
Attribute               Description
======================= ================================================
:attr:`dimension_sizes` The size of each dimension.

:attr:`dimensions`      The dimensions belonging to each grid component.

:attr:`nc_dimensions`   The netCDF dimension name for each dimension
                        in the `dimension_sizes` attribute. If
                        present when the grid is written to a netCDF
                        file, then used for output netCDF dimension
                        names.

:attr:`transform`       A dictionary of :class:`~cf.Transform` objects.
======================= ================================================

**Methods and attributes defined here:**
'''

    def __init__(self, *args, **kwargs):
        '''

Set the built-in dictionary from the arguments and initialize other
attributes.
'''
        # Create the dimension dictionaries
        self.dimension_sizes = {}
        self.dimensions      = {}

        super(Grid, self).__init__(*args, **kwargs)
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''
        return '<CF %s: %s>' % (self.__class__.__name__, 
                                tuple(self.dimension_sizes.values()))
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''
        def _print_coord(grid, key, dimension_coord):
            '''Private function called by __str__'''
            try: 
                grid[key]
            except KeyError: 
                size = grid.dimension_sizes[key]
                try:
                    name = grid.nc_dimensions[key]
                except (KeyError, AttributeError):
                    name = key
                return '%(name)s(%(size)d)' % locals()

            # Still here?
            if dimension_coord:
                # Dimension coordinate
                shape = grid[key].shape
                if shape:
                    shape = '(%d)' % grid[key].size
                else:
                    shape = ''
                name = grid[key].name(ncvar=True, default=key)
            else:
                # Auxiliary coordinate
                try:
                    name = grid[key].name(ncvar=True, default=key)
                except TypeError:
                    name = key
                shape = grid.dimensions[key][:]
                for i in xrange(len(shape)):
                    try:
                        shape[i] = grid[shape[i]].name(ncvar=True, default=shape[i])
                    except KeyError:
                        shape[i] = grid.nc_dimensions[shape[i]]
                shape = str(tuple(shape)).replace("'", "")
                shape = shape.replace(',)', ')')
            #--- End: if
            x = [name]
            x.append(shape)

            try:
                grid[key].compress
            except AttributeError:
                units = getattr(grid[key], 'units', '')
                if units == '1':
                    units = ''

                first = grid[key].first_datum
                last  = grid[key].last_datum


                if grid[key].istime():
                    # Time valued coordinate's first and last values
                    if grid[key].size == 1:
                        x.append(' -> %(first)s' % locals())
                    else:
                        x.append(' -> %(first)s to %(last)s' % locals())
                elif grid[key].dtype.kind == 'S':
                    # String valued coordinate's first and last values
                    if grid[key].size == 1:
                        x.append(" -> '%(first)s' %(units)s" % locals())
                    else:
                        x.append(" -> '%(first)s' to '%(last)s' %(units)s" % locals())
                else:
                    # Numeric valued coordinate's first and last values
                    if grid[key].size == 1:
                        x.append(' -> %(first).7g %(units)s' % locals())
                    else:
                        x.append(' -> %(first).7g to %(last).7g %(units)s' % locals())

            else:
                x.append(' -> compressed ')
                compressed = []
                for unc in grid[key].compress:
                    shape = str(unc.size)
                    compressed.append(unc.name(ncvar=True,
                                               default='unc')+'('+shape+')')
                x.append(', '.join(compressed))

            return ''.join(x)
        #--- End: def

        string = ['Dimensions      : ']
        x = []
        for dim in sorted(self.dimension_sizes):
            x.append(_print_coord(self, dim, True))
        string.append('\n                : '.join(x))
        string.append('\n')

        string.append('Auxiliary coords: ')
        x = []
        for aux in sorted(self.get_keys('^aux')):
            x.append(_print_coord(self, aux, False))
        string.append('\n                : '.join(x))
        string.append('\n')

        # Cell measures
        cell_measures = self.get_keys('^cm')
        if cell_measures:
            string.append('Cell measures   : ')
            x = []
            for cm in sorted(cell_measures):
                x.append(_print_coord(self, cm, False))
            string.append('\n                : '.join(x))
            string.append('\n')

        # Virtual auxiliary coordinates
        try:
            transforms = self.transform.values()
        except AttributeError:
            pass
        else:
            string.append('Virtual coords  : ')
            x = []
            for transform in transforms:
                try:
                    transform['grid_mapping_name']
                    continue
                except KeyError:
                    x.append(repr(transform))
            #-- End: for
            string.append('\n                : '.join(x))

        return ''.join(string)
    #--- End: def

    def dump(self, id=None, omit=()):
        '''
        
Return a string containing a full description of the grid.

:keyword id: *Optional*. Set the common prefix of variable component
             names. If None then defaults to the class name.

:keyword id: str

:keyword omit: *Optional*. Omit the given attributes the description.

:type omit: sequence

:returns: A string containing the description of the grid.

.. seealso:: :func:`cf.dump`
'''
        if id is None:
            id = self.__class__.__name__

        try:
            string = ['%s grid' % self.name]
        except AttributeError:
            string = ['grid']
        string.append(''.ljust(len(string[0]), '-'))

        string.append('%s.dimension_sizes = %s\n' % (id, self.dimension_sizes))
        
        for key in sorted(self.dimensions):
            dims = self.dimensions[key]
            string.append("%(id)s.dimensions['%(key)s'] = %(dims)s" % locals())

        string.append('')

        for key in sorted(self.get_keys('^dim')) + \
                sorted(self.get_keys('^aux')) + \
                sorted(self.get_keys('^cm')):
            string.append(self[key].dump(id="%(id)s['%(key)s']" % locals()))
           
        try:
            for key, transform in self.transform.iteritems():
                string.append(transform.dump( \
                    id="%(id)s.transform['%(key)s']" % locals()))
        except AttributeError:
            pass               
        
        return '\n'.join(string)
    #--- End: def

    def equals(self, other, rtol=None, atol=None):
        '''

Return True if two instances are congruent in that

    1. For each group of grid component types (dimension coordinate,
       auxiliary coordinate, cell measures, etc.) there are an equal
       number of keys and each key's value equals a value in the other
       grid. Note that within a group of component types, the key
       names need not be the same.
  
    2. Each dimension in one grid has an equally sized dimension in
       the other grid.

    3. The unordered set of transforms in one grid equals that of the
       other grid. Noe that The keys of the `transform` attribute
       dictionary need not be the same.

Equality of numbers is to within a tolerance. Refer to :mod:`cf` for
details.

:argument other: The variable to compare against for equality.

:type other: object
             
                                                 
:keyword atol: *Optional*. If None then use the default method for
               setting the absolute tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the absolute tolerance to this value for all such
               comparisons.

:type atol: None or float

:keyword rtol: *Optional*. If None then use the default method for
               setting the relative tolerance for numerical equality
               (refer to :mod:`cf` for details). If a number then set
               the relative tolerance to this value for all such
               comparisons.

:type rtol: None or float

:returns: True if the two objects are congruent, False otherwise.

    '''
        if id(self) == id(other):
            return True
        
        # Check that each instance is the same type
        if type(self) != type(other):
            return False

        try:
            if set(self.dimension_sizes.values()) != \
                    set(other.dimension_sizes.values()):
                return False
        except AttributeError:
            pass

               
        if rtol is None:
            rtol = RTOL()
        if atol is None:
            atol = ATOL()              

        try:
            transform0 = self.transform
        except AttributeError:
            try:
                other.transform
                return False
            except AttributeError:
                pass
        else:
            try:
                transform1 = other.transform
            except AttributeError:
                return False

            self_keys  = sorted(transform0.keys())
            other_keys = sorted(transform1.keys())
            for key0 in self_keys:
                found_match = False
                for key1 in other_keys: 
                    if equals(transform0[key0], transform1[key1],
                              rtol=rtol, atol=atol):
                        found_match = True
                        other_keys.remove(key1)
                        break
                #--- End: for
                if not found_match:
                    return False
            #--- End: for
        #--- End: try

        for coord_type in ('^dim', '^aux', '^cm'):
            self_keys  = sorted(self.get_keys(coord_type))
            other_keys = sorted(other.get_keys(coord_type))
            for coord0 in self_keys:
                found_match = False
                for coord1 in other_keys: 
                    if equals(self[coord0], other[coord1], 
                              rtol=rtol, atol=atol):
                        found_match = True
                        other_keys.remove(coord1)
                        break
                #--- End: for
                if not found_match:
                    return False
            #--- End: for
        #--- End: for

        return True
    #--- End: def

    def coord(self, name, role=None, key=False, exact=False, 
              axis=False, ncvar=None, long=None):
        '''

Find a coordinate of the grid by name. Refer to :meth:`cf.Space.coord`
for details.

'''       
        dim, aux = True, True
        if role == 'dim':
            aux = False
        elif role == 'aux':
            dim = False
                    
        if name is None:
            return None

        found_name = 0

        if dim:
            dim_list = self.get_keys('^dim')
        if aux:
            aux_list = self.get_keys('^aux')       

        if exact:
            if dim:
                for dim in dim_list:
                    if name == self[dim].name(ncvar=True,
                                              default=None):
                        found_name += 1
                        grid_key = dim
                #--- End: for
            #--- End: if            
                    
            if found_name > 1:
                return None
        
            if aux:
                for aux in aux_list:
                    if name == self[aux].name(ncvar=True,
                                              default=None):
                        if found_name:
                            return None
                        grid_key = aux
            #--- End: if

            if found_name != 1:
                return None

            try:
                return self[grid_key]            
            except KeyError:
                return None
        #--- End: if

        # Still here? Look for partial matches            
        get = []
        if dim:
            get.extend(dim_list)
        if aux:
            get.extend(aux_list)
        for key1 in get:
            full = self[key1].name(ncvar=True, default=None)
            if full is None: 
                continue
            if full.startswith(name):
                # e.g. name='lati'
                grid_key    = key1
                found_name += 1                

        if found_name != 1:
            return None

        if key:
            return grid_key

        return self[grid_key]
    #--- End: def
#--- End: clas


class Transform(CfDict):
    '''

A CF transform object defining a coordinate's transformation in a
dictionary-like object.

The named parameters and their values of the transformation (i.e. the
transformation's mappings) comprise the object's key-value pairs.

A transformation is equivalent to either a netCDF(CF) 'formula_terms'
or 'grid_mapping' property. The latter is identified by the presence
of the 'grid_mapping_name' key which contains the mapping's name.

In the 'formula_terms' case, a mapping to a coordinate (as opposed to
another space) uses the coordinate's grid key name as a pointer rather
than a copy of the coordinate itself.

Refer to the `Grid Structure` section of the :mod:`cf` module for
details.

:argument *args: Keys and values are initialized exactly as for a
                 built-in dict.

:argument **kwargs: Keys and values are initialized exactly as for a
                    built-in dict.


**Overloaded operators:**

The ``in`` (set membership) operator is overloaded to use numerically
tolerant equality.


**Special attributes:**

============ =======================================================
Attribute    Description
============ =======================================================
:attr:`name` The identifying name of the transformation. Typically a
             netCDF(CF) 'formula_terms' or 'grid_mapping_name'.
============ =======================================================


**Examples:**

>>> t
<CF Transform: atmosphere_sigma_coordinate>
>>> dump(t)
atmosphere_sigma_coordinate transform
-------------------------------------
Transform['ps'] = <CF Space: surface_air_pressure(73, 96)>
Transform['ptop'] = 0.05
Transform['sigma'] = 'dim0'

>>> t
<CF Transform: rotated_latitude_longitude>
>>> dump(t)
rotated_latitude_longitude transform
------------------------------------
Transform['grid_mapping_name'] = 'rotated_latitude_longitude'
Transform['grid_north_pole_latitude'] = 33.67
Transform['grid_north_pole_longitude'] = 190.0

**Methods and attributes defined here:**
'''
    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
''' 
        try:
            return '<CF %s: %s>' % (self.__class__.__name__, self.name)
        except AttributeError:
            return '<CF %s: >' % self.__class__.__name__
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''    
        return 'Transform       : %s' % repr(self)
    #--- End: def

    def dump(self, id=None, omit=()):
        '''

Return a string containing a full description of the transformation.

:keyword id: *Optional*. Set the common prefix of component names. If
             None then defaults to the class name.

:type id: str

:keyword omit: *Optional*. Omit the given attributes from the
               description.

:type omit: sequence


:returns:  A string containing the description of the object.

.. seealso:: :func:`cf.dump`
'''
        if id is None:
            id = self.__class__.__name__
            
        try:
            string = ['%s transform' % self.name]
        except AttributeError:
            string = [' transform']

        string.append(''.ljust(len(string[0]), '-'))

        for key in sorted(self.keys()):
            string.append("%s['%s'] = %s" % (id, key, repr(self[key])))

        return '\n'.join(string)
    #--- End: def

#--- End: class


class CellMethods(CfList):
    '''

A CF cell methods object to describe the characteristic of a field
that is represented by cell values.

Each cell method is stored in a dictionary and these dictionaries are
stored in a list-like object. Similarly to a CF 'cell_methods' string,
the order of cell methods in the list is important.

The dictionary representing each cell method recognizes the following
keys (where all keys are optional; the referenced sections are from
`NetCDF Climate and Forecast (CF) Metadata Conventions Version 1.5
<http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.5/cf-conventions.html>`_;
and words in double quotes ("...") refer to CF cell_methods components
in the referenced sections):

========= ========================================================
Key       Value
========= ========================================================
method    A "method" component. Refer to section 7.3.

name      A list of all of the "name" components involved with the
          "method" component. Each element is either:

              1. The standard_name of the cell method's dimension
              2. The string 'area'          
              3. A netCDF variable name or netCDF dimension name

          Option 3. only occurs if the cell method's dimension
          either has no dimension coordinate or has a dimension
          coordinate with no standard_name.

          This is a deviation from the "name" components as
          described in sections 7.3, 7.3.1 and 7.3.4, in that
          standard_names are stored whenever possible, rather than
          only in the special cases described in section
          7.3.4. Inspection of the 'no_coords' list retains the
          information required to fully interpret the cell method.

dim       A list of space dimension names corresponding to the
          'name' list. A 'name' list value of 'area' always
          corresponds to a 'dim' list value of None.

no_coords A list of booleans, corresponding to the 'name' list,
          indicating whether or not a particular cell method is
          relevant to the data in a way which may not be precisely
          defined by the corresponding dimension or dimensions. Refer
          to section 7.3.4.

interval  A list of "interval" component values corresponding to
          the 'name' list. Refer to section 7.3.2.

units     A list of "interval" component units corresponding to
          the 'interval' list. Refer to section 7.3.2.

comment   A "comment" or non-standardised information,
          component. Refer to sections 7.3.2 and 7.4

within    A "within" climatology component. Refer to section 7.4.

over      An "over" cell portion or "over" climatology component.
          Refer to sections 7.3.3 and 7.4.

where     A "where" component. Refer to section 7.3.3.
========= ========================================================

Note that the above table assumes that the cell methods have been
constructed in the context of a space, the only way in which the 'dim'
and 'no_coords' keys may take sensible values. If this is not the
case, as would happen if the `parse` method were called without its
`space` keyword, then elements of the 'dim' key default to False and
elements of the 'no_coords' key default to None.

:argument sequence: *Optional*. Initialize new list from
                    sequence's items.

:type sequence: sequence

**Examples:**

>>> c = cf.CellMethods()
>>> c = c.parse('time: minimum within years time: mean over years (ENSO years)')    
>>> print c
Cell methods    : time: minimum within years
                  time: mean over years (ENSO years)
>>> list(c)
[
 {'method'   : 'minimum',
  'name'     : ['time'],     
  'dim'      : [False],
  'no_coords': [None], 
  'within'   : 'years'
 },
 {'method'   : 'mean'
  'name'     : ['time'],
  'dim'      : [False], 
  'no_coords': [None],     
  'comment'  : 'ENSO years', 
  'over'     : 'years'
 }
]

>>> c = c.parse('lat: lon: standard_deviation')
>>> list(c)
[
 {'dim'      : [False, False],
  'method'   : 'standard_deviation'
  'name'     : ['lat', 'lon'],
  'no_coords': [None, None],
 }
]
 

**Overloaded operators**

Refer to :class:`~cf.CfList`.


**Methods:**

================================== =====================================================
Method                             Description
================================== =====================================================
:meth:`~cf.CellMethods.append`     As for a built-in list.
:meth:`~cf.CellMethods.copy`       Create a deep copy.
:meth:`~cf.CellMethods.count`      As for a built-in list but using numerically tolerant
                                   equality.
:meth:`~cf.CellMethods.dump`       Return a string containing a full description of the
                                   object.
:meth:`~cf.CellMethods.equals`     Determine whether two lists are congruent
                                   element-wise.
:meth:`~cf.CellMethods.extend`     As for a built-in list.
:meth:`~cf.CellMethods.index`      As for a built-in list but using numerically tolerant
                                   equality.
:meth:`~cf.CellMethods.insert`     As for a built-in list.
:meth:`~cf.CellMethods.parse`      Parse a netCDF(CF) cell_methods string.
:meth:`~cf.CellMethods.pop`        As for a built-in list.
:meth:`~cf.CellMethods.properties` Return a set of writable, public attributes.
:meth:`~cf.CellMethods.remove`     As for a built-in list.
:meth:`~cf.CellMethods.reverse`    As for a built-in list.
:meth:`~cf.CellMethods.sort`       As for a built-in list.
:meth:`~cf.CellMethods.strings`    Convert each element to a string.
================================== =====================================================


**Methods and attributes defined here:**
'''

#    def __init__(self, string=None, space=None):
#        self.parse(string=string, space=space)
#    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''
        return '<CF %s: %s>' % (self.__class__.__name__, 
                                ' '.join(self.strings()))
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''        
        strings = self.strings()

        try:
            string = ['Cell methods    : %s' % strings[0]]
        except IndexError:
            return  'Cell methods    : '

        for s in strings[1:]:
            string.append('                  %s' % s)
            
        return '\n'.join(string)
    #--- End: def

    def dump(self, id=None, omit=()):
         '''

Return a string containing a full description of the cell methods.

If a cell methods 'name' is followed by a '*' then that cell method is
relevant to the data in a way which may not be precisely defined its
corresponding dimension or dimensions.

:keyword id: *Optional*. Set the common prefix of each line of the
             output. If None then defaults to the class name.

:type id: str

:keyword omit: *Optional*. Dummy argument required for consistency
               with the :func:`cf.dump` function.

:type omit: sequence


:returns: A string containing the description of the cell methods.

.. seealso:: :func:`cf.dump`
'''
         if id is None:
             id = self.__class__.__name__
                              
         string = [] #'cell methods\n------------']
         
         for i, s in enumerate(self.strings()):
             string.append('%s[%d] -> %s' % (id, i, s))
             
         return '\n'.join(string)
    #--- End: def

    def parse(self, string=None, space=None):
        '''

Parse a CF cell_methods string into this CellMethods instance.

:argument string: The CF cell_methods string to be parsed into the
                  CellMethods object.

:type string: str

:keyword space: *Optional*. Set the 'name', 'dim' and 'no_coord' key
                values appropriately for the given space. Refer to
                :class:`~cf.CellMethods`.

:type space: Space

:returns: A cell methods list.

**Examples:**

>>> c = cf.CellMethods()
>>> c = c.parse('time: minimum within years time: mean over years (ENSO years)')    
>>> print c
Cell methods    : time: minimum within years
                  time: mean over years (ENSO years)

'''

        result = type(self)()

        # Split the cell_methods string into a list of strings ready
        # for parsing into the result list. E.g.
        #   'lat: mean (interval: 1 hour)'
        # maps to 
        #   ['lat:', 'mean', '(', 'interval:', '1', 'hour', ')']
        cell_methods = re.sub('\((?=[^\s])' , '( ', string)
        cell_methods = re.sub('(?<=[^\s])\)', ' )', cell_methods).split()

        while cell_methods:

            # Create a new element for this 
            result.append(CfDict())            

            # List of names
            result[-1]['name']      = CfList()
            result[-1]['dim']       = CfList()
            result[-1]['no_coords'] = CfList()

            while cell_methods:
                if not cell_methods[0].endswith(':'):
                    break
                result[-1]['name'].append(cell_methods.pop(0)[:-1])            
                result[-1]['dim'].append(False)
                result[-1]['no_coords'].append(None)
                            
            if not cell_methods:
                break

            # Method
            result[-1]['method'] = cell_methods.pop(0)

            if not cell_methods:
                break

            # Climatological statistics and statistics which apply to
            # portions of cells
            while cell_methods[0] in ('within', 'where', 'over'):
                term = cell_methods.pop(0)
#                if term == 'over' and cell_methods[0] in ('years', 'days'):
#                    result[-1]['c'+term] = cell_methods.pop(0)
#                else:
                result[-1][term] = cell_methods.pop(0)
                if not cell_methods:
                    break
            #--- End: while 

            if not cell_methods: 
                break

            # interval and comment
            if cell_methods[0].endswith('('):
                cell_methods.pop(0)

                if not (re.search('^(interval|comment):$', cell_methods[0])):
                    cell_methods.insert(0, 'comment:')
                           
                while not re.search('^\)$', cell_methods[0]):
                    term = cell_methods.pop(0)[:-1]

                    if term == 'interval':
                        if not result[-1].has_key('interval'):
                            result[-1]['interval'] = CfList()
                            result[-1]['units']    = CfList()
                        interval = cell_methods.pop(0)
                        try:
                            interval = float(interval)
                        except ValueError:
                            pass
                        result[-1]['interval'].append(interval)
                        if cell_methods[0] != ')':
                            result[-1]['units'].append(cell_methods.pop(0))
                        else:
                            result[-1]['units'].append('')
                        #--- End: if
                        continue
                    #--- End: if

                    if term == 'comment':
                        comment = CfList()
                        while cell_methods:
                            if cell_methods[0].endswith(')'):
                                break
                            if cell_methods[0].endswith(':'):
                                break
                            comment.append(cell_methods.pop(0))
                        result[-1]['comment'] = ' '.join(comment)
                    #--- End: if

                #--- End: while 

                if cell_methods[0].endswith(')'):
                    cell_methods.pop(0)
            #--- End: if
        #--- End: while

        # ----------------------------------------------------------
        # If no space has been provided, then return now without
        # adjusting the 'name', 'dim' and 'no_coords' keys
        # ----------------------------------------------------------
        if space is None:
            return result

        # ----------------------------------------------------------
        # Still here? Then adjust the 'name', 'dim' and 'no_coords'
        # keys according to the given space.
        # ----------------------------------------------------------
        try: 
            space.grid.nc_dimensions
        except AttributeError:
            return result

        # Change each 'name' value to a standard_name (or grid
        # coordinate key) and create the 'dim' key
            
        # From the CF conventions (1.5): In the specification of this
        # attribute, name can be a dimension of the variable, a scalar
        # coordinate variable, a valid standard name, or the word
        # 'area'.
        for j in xrange(len(result)):
            for i in xrange(len(result[j]['name'])):

                # Since input scalar coordinate variables have now
                # been converted into 1d dimension coordinates, name
                # being "a dimension of the variable, a scalar
                # coordinate variable" may be checked by testing name
                # against space.grid.nc_dimensions.values().                   
                found_ncdimension = False
                for dim, ncdim in space.grid.nc_dimensions.iteritems():
                    if result[j]['name'][i] == ncdim:
                        result[j]['no_coords'][i] = False
                        result[j]['dim'][i]       = dim
                        try:                            
                            result[j]['name'][i] = space.grid[dim].name(default=ncdim)
                        except KeyError:
                            result[j]['name'][i] = ncdim
                        found_ncdimension = True
                        break

                if found_ncdimension:
                    continue

                # The cell_methods "name" component is either a
                # standard_name or the string 'area'
                result[j]['no_coords'][i] = True
                if result[j]['name'][i] == 'area':
                    result[j]['dim'][i] = None
                else:
                    for key in space.grid.get_keys('dim|aux'):
                        try:
                            sn = space.grid[key].standard_name
                            if result[j]['name'][i] == sn:
                                dim = space.grid.coord(sn, key=True, exact=True)
                                result[j]['dim'][i] = dim
                                break
                        except AttributeError:
                            pass
                    #--- End: for

#                        sn = space.grid[key].getpub('standard_name', None)
#                        if result[j]['name'][i] == sn:
#                            dim = space.grid.coord(sn, key=True, exact=True)
#                            result[j]['dim'][i] = dim
#                            break

        #--- End: for

        return result
    #--- End: def

    def strings(self):
        '''

Retun a list of a netCDF(CF)-like string representation of each cell
method.

Note that if the intention is to concatenate the output list into a
string for creating a netCDF(CF) cell_methods attribute, then the cell
methods "name" components may need to be modified, where appropriate,
to reflect netCDF variable names.

:returns: A list of cell method strings.

**Examples:**

>>> c = cf.CellMethods()
>>> c = c.parse('time: minimum within years time: mean over years (ENSO years)')    
>>> print c
Cell methods    : time: minimum within years
                  time: mean over years (ENSO years)
>>> c.strings()
['time: minimum within years',
 'time: mean over years (ENSO years)']
'''
        strings = []
        
        for cm in self:

            string = []

            for j in xrange(len(cm['name'])):
                string.append('%s: ' % cm['name'][j])

            string.append(cm['method'])

            x = []
            try:
                x.extend(('within', cm['within'])) 
            except KeyError:
                try:
                    x.extend(('where', cm['where'])) 
                except KeyError:
                    pass
            try:
                x.extend(('over', cm['over'])) 
            except KeyError:
                pass
            if x:
                string.append(' %s' % ' '.join(x))

            if cm.has_key('interval'):
                string.append(' (')
                for j in xrange(len(cm['interval'])):
                    string.append('interval: %s' % repr(cm['interval'][j]))
                    try:
                        string.append(' %s' % cm['units'][j])
                    except KeyError:
                        pass
                if cm.has_key('comment'):
                    string.append(' comment: %s' % cm['comment'])
                string.append(')')

            elif cm.has_key('comment'):
                string.append(' (%s)' % cm['comment'])
            
            strings.append(''.join(string))
        #--- End: for

        return strings
    #--- End: defstrings

#--- End: class


# ====================================================================
#
# Variable list objects
#
# ====================================================================

class VariableList(CfList):
    '''

An ordered sequence of variables stored in a single, list-like
object. Refer to Variable Lists in the :mod:`cf` package for details.

**Methods and attributes defined here:**
'''

    _private_attrs = Variable._private_attrs[:]
        
    def __getattribute__(self, attr, *default):
        '''
x.__getattribute__(attr) <==> x.attr     
'''
        if (attr.startswith('__') or 
            attr in super(VariableList, self).__getattribute__('__dict__') or
            attr in dir(self)):
            # Return an attribute of the list-like object
            return super(VariableList, self).__getattribute__(attr, *default)

        # Return a built-in list of the attributes from each element
        # of the cf list
        return [getattr(v, attr, *default) 
                for v in super(VariableList, self).__getattribute__('_list')]
    #--- End: def

    def __setattr__(self, attr, value):
        '''
x.__setattr__(attr, value) <==> x.attr=value
'''
        if (attr.startswith('__') or 
            attr in super(VariableList, self).__getattribute__('__dict__') or
            attr in dir(self)):
            # Set an attribute on the list-like object
            super(VariableList, self).__setattr__(attr, value)
            return
        #--- End: if

        # Still here? Then do a built-in setattr on each of the list's
        # elements
        for v in self.__dict__['_list']:
            setattr(v, attr, value)
    #--- End: def

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)
'''
        return repr(self._list).replace('>, ', '>,\n ')
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)
'''
        return '\n\n'.join([v.__str__() for v in self._list])
    #--- End: def

    def __eq__(self, other):
        '''
x.__eq__(y) <==> x==y
'''
        return type(self)([v.__eq__(other) for v in self._list])
    #--- End: def

    def __ne__(self, other):
        '''
x.__ne__(y) <==> x!=y
'''
        return type(self)([v.__ne__(other) for v in self._list])
    #--- End: def

    def __neg__(self):
        '''
x.__neg__() <==> -x
'''
        return type(self)([v.__neg__(other) for v in self._list])
    #--- End: def

    def __pos__(self):
        '''
x.__pos__() <==> +x
'''
        return type(self)([v.__pos__(other) for v in self._list])
    #--- End: def

    def __abs__(self):
         '''
x.__abs__() <==> abs(x)
'''
         return type(self)([v.__abs__(other) for v in self._list])
    #--- End: def
    
    def __invert__(self):
         '''
x.__invert__() <==> ~x
'''
         return type(self)([v.__invert__(other) for v in self._list])
    #--- End: def
    
    def __add__(self, other):
        '''
x.__add__(y) <==> x+y
        '''
        return type(self)([v.__add__(other) for v in self._list])
    #--- End: def

    def __sub__(self, other):
        '''
x.__sub__(y) <==> x-y
        '''
        return type(self)([v.__sub__(other) for v in self._list])
    #--- End: def

    def __mul__(self, other):
        '''
x.__mul__(y) <==> x*y
        '''
        return type(self)([v.__mul__(other) for v in self._list])
    #--- End: def

    def __div__(self, other):
        '''
x.__div__(y) <==> x/y
'''
        return type(self)([v.__div__(other) for v in self._list])
    #--- End: def

    def __radd__(self, other):
        '''
x.__radd__(y) <==> y+x
'''
        return type(self)([v.__radd__(other) for v in self._list])
    #--- End: def

    def __rmul__(self, other):
        '''
x.__rmul__(y) <==> y*x
'''
        return type(self)([v.__rmul__(other) for v in self._list])
    #--- End: def

    def __iadd__(self, other):
        '''
x.__iadd__(y) <==> x+=y
'''
        for v in self._list:
            v.__iadd__(other)
        return self
    #--- End def
        
    def __isub__(self, other):
        '''
x.__isub__(y) <==> x-=y
'''
        for v in self._list:
            v.__isub__(other)
        return self
    #--- End def

    def __imul__(self, other):
        '''
x.__imul__(y) <==> x*=y
'''
        for v in self._list:
            v.__imul__(other)
        return self
    #--- End def

    def __idiv__(self, other):
        '''
x.__idiv__(y) <==> x/=y
'''   
        for v in self._list:
            v.__idiv__(other)
        return self
    #--- End def

    @property
    def slice(self):
        '''
 
Slice each variable in the list, returning a new list of variables.

**Examples:**

>>> vl
[<CF Variable: air_temperature(73, 96)>,
 <CF Variable: air_temperature(73, 96)>]
>>> vl.slice[0,0]
[<CF Variable: air_temperature(1,1)>,
 <CF Variable: air_temperature(1,1)>]
'''
        return SliceVariableList(self)
    #--- End: def

    def _concatenate_data(self):
        '''

'''
        for v in self._list:
            v._concatenate_data()
    #--- End: def    

    def dump(self, *arg, **kwargs):
        '''

Return a string containing the full descriptions of each variable in
the list.

.. seealso:: :func:`cf.dump`
'''
        return '\n\n\n'.join([v.dump(*arg, **kwargs) for v in self._list])
    #--- End: def

    def type(self):
        '''

The types of the data stored in each variable's Data object.

:returns: A list of list of type objects for each of the data.
elements.

 **Examples:**

>>> v.type()
[[<type 'netCDF4.Variable'>]]
>>> v.slice[...]
>>> v.type()
[[<type 'numpy.ndarray'>]]
'''
        return [v.type() for v in self._list]
    #--- End: def

    def pub(self, *args, **kwargs):
        '''

Set a public attribute from each element of the list of variables or
return a built-in list of the public attributes of each element of the
list of variables.
'''
        if kwargs:
            [v.pub(*args, **kwargs) for v in self._list]
        else:
            return [v.pub(*args) for v in self._list]
    #--- End: def

    def setpub(self, attr, value):
        '''

Set a public attribute from each element of the list of variables.
'''
        for v in self._list:
            v.setpub(attr, value)
    #--- End: def

    def delpub(self, attr):
        '''

Delete a public attribute from each element of the list of variables.
'''
        for v in self._list:
            v.delpub(attr)
    #--- End: def

    def getpub(self, *args):
        '''

Return a built-in list of the public attributes of each element of the
list of variables.
'''
        return [v.getpub(*args) for v in self._list]
    #--- End: def

    def priv(self, *args, **kwargs):
        '''

Set a private attribute from each element of the list of variables or
return a built-in list of the private attributes of each element of
the list of variables.
'''
        if kwargs:
            [v.priv(*args, **kwargs) for v in self._list]
        else:
            return [v.priv(*args) for v in self._list]
    #--- End: def

    def setpriv(self, attr, value):
        '''

Set a private attribute on each element of the list of variables.
'''
        for v in self._list:
            v.setpriv(attr, value)
    #--- End: def

    def delpriv(self, attr):
        '''

Delete a private attribute from each element of the list of variables.
'''
        for v in self._list:
            v.delpriv(attr)
    #--- End: def

    def getpriv(self, attr, *default):
        '''

Return a built-in list of the private attributes of each element of the
list of variables.
'''
        return [v.getpriv(attr, *default) for v in self._list]
    #--- End: def

    def name(self, *arg, **kwargs):
        '''

Return a built-in list of the names of each element of the list of
variables.
'''
        return [v.name(*arg, **kwargs) for v in self._list]
    #--- End: def

    def match(self, *arg, **kwargs):        
        '''
 
Return a built-in list of the result of the match on each element of
the list of variables, with unambiguous abbreviations of phenomenon
names allowed.
'''
        return [v.match(*arg, **kwargs) for v in self._list]
    #--- End: def
       
    def ematch(self, *arg, **kwargs):
        '''

Return a built-in list of the result of the match on each element of
the list of variables, with only exact phenomenon names allowed.
'''
        return [v.ematch(*arg, **kwargs) for v in self._list]
    #--- End: def
       
    def extract(self, *arg, **kwargs):
        '''

Return the list of variables which match the given phenomena, with
unambiguous abbreviations of phenomenon names allowed.
'''

        return type(self)([v for v in self._list if v.match(*arg, **kwargs)])
    #--- End: def
  
    def eextract(self, *arg, **kwargs):
        '''

Return the list of variables which match the given phenomena, with
only exact phenomenon names allowed.
'''
        return type(self)([v for v in self._list if v.ematch(*arg, **kwargs)])
    #--- End: def
  
#--- End: class


class CoordinateList(VariableList):
    '''

An ordered sequence of coordinates stored in a single, list-like
object. Refer to Variable Lists in the :mod:`cf` package for details.

**Methods and attributes defined here:**
'''
    pass
#--- End: class


class SpaceList(VariableList):
    '''

An ordered sequence of spaces stored in a single, list-like
object. Refer to Variable Lists in the :mod:`cf` package for details.

**Methods and attributes defined here:**
'''
    @property
    def slice(self):
        '''

Slice each space in the list, returning a new list of spaces. Slicing
by indices or by coordinate values are both allowed.

**Examples:**

>>> sl
[<CF Variable: air_temperature(73, 96)>,
 <CF Variable: air_temperature(73, 96)>]
>>> sl.slice[0,0]
[<CF Variable: air_temperature(1,1)>,
 <CF Variable: air_temperature(1,1)>]
>>> sl.slice(longitude=0, latitude=0)
[<CF Variable: air_temperature(1,1)>,
 <CF Variable: air_temperature(1,1)>]
'''
        return SliceSpaceList(self)
    #--- End: def

    def coord(self, *args, **kwargs):
        '''

Return the requested coordinate from each space in the list.
'''    
        return CoordinateList([v.coord(*args, **kwargs) for v in self._list])
    #--- End: def
    
    def squash(self):
        '''
 
Return a built-in list of the result of the match on each element of
the list of variables, with unambiguous abbreviations of phenomenon
names allowed.
'''  
        for f in self._list:
            f.squash()
    #--- End: def
       
#--- End: class


# ====================================================================
#
# Slice objects
#
# ====================================================================
class SliceVariable(object):

    __slots__ = ['variable']

    def __init__(self, variable):
        '''
        
Set the contained variable.
'''
        self.variable = variable
    #--- End: def

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
'''
        variable  = self.variable
        new       = variable.copy(data=False)
        new._data = variable._data.slice[index]

        return new
   #--- End: def

    def __setitem__(self, index, value):
        '''
x.__setitem__(index, value) <==> x[index] = value
'''
        self.variable._data.slice[index] = value
    #--- End: def

#--- End: class


class SliceCoordinate(SliceVariable):

    __slots__ = []

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
'''
        coord = self.variable

        # Copy the coordinate, omitting the data
        #dch but the bounds/climatology are still deep copied .....
        new = coord.copy(data=False)

        # Parse the index
        index, dummy = _parse_index(coord, index)

        try:
            new._data = coord._data.slice[index]
        except AttributeError:
            pass

        # Slice the bounds, if there are any
        try:
            new.bounds = coord.bounds.slice[index+(slice(None),)]
        except AttributeError:
            try:
                new.climatology = coord.climatology.slice[index+(slice(None),)]
            except AttributeError:
                pass
      
        # Return the new coordinate
        return new
    #--- End: def

#--- End: class

class SliceSpace(SliceVariable):
    '''

Return a new Space instance whose data and grid are sliced according
to indices or call arguments to its `slice` attribute. The latter
converts criteria on coordinate values into indices and so the two
methods a formally equivalent.

**Examples:**

>>> print s
Data            : air_temperature(time, latitude, longitude)
Cell methods    : time: mean (interval: 1.0 month)
Dimensions      : time(7070) -> 450-11-16 00:00:00 to 1049-12-16 12:00:00
                : latitude(64) -> -87.8638 to 87.8638 degrees_north
                : longitude(128) -> 0 to 357.1875 degrees_east
                : height(1) -> 2 m
Auxiliary coords:
>>> s.slice[:,32, 64]
<CF Space: air_temperature(12, 1, 1)
>>> s.slice(longitude=0.0, latitude=0.0)
<CF Space: air_temperature(12, 1, 1)

Slicing by index
----------------
Refer to the 'Copying and resizing' of the `cf` help.

Slicing by coordinate criteria
------------------------------

Coordinates are identified by keyword parameters of their standard
names. If the optional positional argument `exact` evaluates to `False`
(the default) then unambiguous abbreviations of the standard names are
accepted, otherwise exact standard names are required:

>>> s.slice(longitude=0.0)
<CF Space: air_temperature(12, 64, 1)
>>> s.slice(lon=0.0)
<CF Space: air_temperature(12, 64, 1)
>>> s.slice(l=0.0)
ValueError: Can't identify coordinate from keyword 'l'
>>> s.slice(False, lon=0.0)
<CF Space: air_temperature(12, 64, 1)
>>> s.slice(True, lon=0.0)
ValueError: Can't identify coordinate from keyword 'lon'

The coordinate criteria are given by the keyword parameters
values. Four different methods are available:

1. A scalar. The dimension is sliced where coordinate equals (using
   the coordinate's == operator) the scalar value.

   **Example**
   >>> s.slice(lat=0.0)

2. A sequence. The dimension is sliced where the coordinate equals
   (using the coordinate's == operator) the sequence's elements. All
   elements of the sequence must be found in the coordinate, but the
   sequence may contain fewer elements than the coordinate's data.

   **Example**
   >>> s.slice(lon=[0.0, 2.8125])

3. A :class:`~cf.Comparison` object (as, for example, returned by the
   :func:`~cf.lt` function). The dimension is sliced where the
   comparison for the coordinate evaluates to True.

   **Example**
   >>> s.slice(lat=cf.lt(0.0))

4. An numpy array of booleans with the same shape as the
   coordinate. The dimension is sliced where elements of this array
   are True.

   **Example**
   >>> s.slice(lat=s.coord('lat')<0.0)
  
If a coordinate is not identified with a keyword parameter then its
dimension is returned unchanged.

The order in which coordinates are given in the parameter list is
irrelevant.

A call with no keyword parameters is equivalent to indexing with an
Ellipsis (...).

**Examples:**

>>> s.slice(lat=0.0, lon=0.0)
<CF Space: air_temperature(12, 1, 1)
>>> s.slice(lon=cf.lt(0.0), lon=[0, 2.8125])
<CF Space: air_temperature(12, 32, 2)
>>> s.slice(lon=[0, 1])
ValeuError: 'longitude' coordinate doesn't contain the value 1
>>> s.slice(lon=1)
ValueError: No indices found for 'longitude' coordinate
'''
    __slots__ = []

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
''' 
        space = self.variable

        # Copy the field

        # It'd be nice not to deep copy the grid coordinates and cell
        # measures now but do this in the for loop below. Would be a
        # bit more efficient.

        # Copy the coordinate
        new = space.copy(data=False)

        # Parse the index
        index, dummy = _parse_index(space, index)

        # Slice the field's data
        new._data = space._data.slice[index]

        # Break ancillary variables
        full_slice = (slice(None),) * space.ndim
        if index != full_slice:
            try:
                del new.ancillary_variables
            except AttributeError:
                pass        

        try:
            new.grid
        except AttributeError:
            # Return now if the field doesn't have a grid
            return new
        
        # -----------------------------------------------------------
        # Still here? Then slice the space
        # -----------------------------------------------------------
        i = -1
        for dim in new.grid.dimensions['data'][:]:

            i += 1

            # Continue if no slicing to be done on this dimension
            if index[i] == slice(None, None, None): # dch ??
                continue

            for key in new.grid:

                # Continue if this space component does not span the
                # current dimension
                if dim not in new.grid.dimensions[key]:
                    continue

                dice          = [slice(None)] * new.grid[key].ndim
                pos           = new.grid.dimensions[key].index(dim)
                dice[pos]     = index[i]
                new.grid[key] = new.grid[key].slice[tuple(dice)]

                # Update dimension size
                new.grid.dimension_sizes[dim] = new.grid[key].shape[pos]
            #--- End: for
        #--- End: for
                                
        try:
            transforms = new.grid.transform.values()
        except AttributeError:
            pass
        else:
            if index != full_slice:
                # Break formula terms
                for transform in transforms:
                    try:
                        transform['grid_mapping_name']
                        continue
                    except KeyError:
                        pass
                    for term, value in transform.iteritems():
                        if isinstance(value, Variable): 
                            transform[term] = None
            #--- End: if
        #--- End: try

        return new
    #--- End: def

    def __call__(self, *args, **kwargs):
        '''

Slice by coordinate value. refer to :meth:`cf.Space.slice` for
details.
'''

        space = self.variable       

        if not kwargs:
            return copy.deepcopy(space)

        if args:
            exact = args[0]
        else:
            exact = False

        dimensions = []

        # Initialize index
        index  = [slice(None)] * space.ndim

        # Loop round slice criteria
        for name, value in kwargs.iteritems():
                                    
            key = space.coord(name, key=True, exact=exact)

            if key is None:
                raise ValueError, "Can't identify coordinate from keyword '%(name)s'" % locals()

            if key.startswith('dim'):
                # Dimension coordinate
                dim = key
            else:
                # Auxiliary coordinate
                dim = space.grid.dimensions[key][0]

                if space.grid[key].ndim > 1:
                    raise ValueError, "Can't define a slice for a multidimensional auxiliary coordinate '%s'" % space.coord(name, exact=exact)
            #--- End: if

            if dim in dimensions:
                raise ValueError, "Two coordinates share the same dimension"
            dimensions.append(dim)

            if dim not in space.grid.dimensions['data']:
                raise IndexError, "Data doesn't have a '%s' dimension" % \
                    space.coord(name, exact=exact)

            # Define c, the coordinate for the specified dimension
            c = space.grid[key]
            slice_ = []
            try:
                # value is a Comparison object
                slice_ = value.evaluate(c)
            except AttributeError:

                try:
                    # value is a numpy array
                    if value.dtype.kind == 'b':
                        # value is a boolean numpy array
                        if value.shape != c.shape:
                            raise ValueError, "Boolean numpy array must have shape as coordinate"
                        slice_ = value
                    else:
                        # value is a non-boolean numpy array
                        try:
                            # value is a 1d numpy array
                            for v in value:
                                x = numpy.where(c == v)[0]
                                if not x.size:
                                    raise ValueError, "'%s' coordinate doesn't contain the value %s" % \
                                        (space.coord(name, exact=exact), v)
                                slice_ = x
                        except TypeError:
                            # value is a scalar numpy array
                            slice_ = numpy.where(c == v)[0]
                except AttributeError:
                    # value is not a numpy array
                    try:
                        # value is a sequence
                        value.__iter__
                        for v in value:
                            x = numpy.where(c == v)[0]
                            if not x.size:
                                raise ValueError, "'%s' coordinate doesn't contain the value %s" % \
                                    (space.coord(name, exact=exact), v)
                            slice_ = x
                    except AttributeError:
                        # value is a scalar
                        slice_ = numpy.where(c == value)[0]
                    #--- End: try
                #--- End: try
            #--- End: try
                
            new_dim_size = len(slice_)
            if not new_dim_size:
                raise ValueError, "No indices found for '%s' coordinate" % \
                    space.coord(name, exact=exact)
            #--- End: if

            # Add slice_ to the right place in the data slice
            pos        = space.grid.dimensions['data'].index(dim)
            index[pos] = slice_        
        #--- End: for

        return space.slice[tuple(index)]
    #--- End: def

#--- End: class


class SliceVariableList(object):

    __slots__ = ['variablelist']

    def __init__(self, variablelist):
        '''
        
Set the contained variablelist.
'''
        self.variablelist = variablelist
    #--- End: def

    def __getitem__(self, index):
        '''
x.__getitem__(index) <==> x[index]
'''
        variablelist = self.variablelist
        out          = [v.slice[index] for v in variablelist]
        return type(variablelist)(out)
    #--- End: def

    def __setitem__(self, index, value):
        '''
x.__setitem__(index, value) <==> x[index]=value
'''
        for v in self.variablelist:
            v.slice[index] = value
    #--- End: def

#--- End: class

class SliceSpaceList(SliceVariableList):

    __slots__ = []

    def __call__(self, *arg, **kwargs):
        '''

Slice by spaces by coordinate value. refer to :meth:`cf.Space.slice`
for details.
'''
        variablelist = self.variablelist
        return type(variablelist)([f.slice(*arg, **kwargs) for f in variablelist])
    #--- End: def

#--- End: class
